import base64
import concurrent
import copy
import datetime as dt
import functools
import ipaddress
import json
import logging
import math
import os
import pickle
import random
import re
import shlex
import time
import traceback
import typing as t
import uuid
from collections import OrderedDict

from flask import request, current_app, g
from flask_jwt_extended import jwt_required, get_jwt_identity, create_access_token
from pkg_resources import parse_version

import dimensigon
import dimensigon.use_cases.cluster
import dimensigon.web.network as ntwrk
from dimensigon import defaults as d, defaults
from dimensigon.domain.entities import Software, Server, SoftwareServerAssociation, Catalog, Route, StepExecution, \
    Orchestration, OrchExecution, User, ActionTemplate, ActionType, Vault
from dimensigon.use_cases.deployment import deploy_orchestration, validate_input_chain
from dimensigon.use_cases.use_cases import async_send_file
from dimensigon.utils import asyncio, subprocess
from dimensigon.utils.dag import DAG
from dimensigon.utils.event_handler import Event
from dimensigon.utils.helpers import get_distributed_entities, is_iterable_not_string, md5, get_now, format_exception
from dimensigon.utils.var_context import Context
from dimensigon.web import db, executor, errors, threading
from dimensigon.web.api_1_0 import api_bp
from dimensigon.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog, log_time
from dimensigon.web.helpers import check_param_in_uri, normalize_hosts, search
from dimensigon.web.json_schemas import launch_command_post, routes_post, routes_patch, \
    launch_orchestration_post, send_post, orchestration_full, manager_server_ignore_lock_post

if t.TYPE_CHECKING:
    from dimensigon.use_cases.operations import IOperationEncapsulation, CompletedProcess


@api_bp.route('/')
def home():
    return "API v1.0 documentation page"


@api_bp.route('/join/token', methods=['GET'])
@jwt_required()
def join_token():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        user_join = User.get_by_name('join')
        expire_time = request.args.get('expires_time', type=int, default=None)

        kwargs = {}
        if expire_time:
            kwargs['expires_delta'] = dt.timedelta(minutes=expire_time)

        return {'token': create_access_token(str(user_join.id), **kwargs)}, 200
    else:
        return {}, 401


@api_bp.route('/join/public', methods=['GET'])
@jwt_required()
def join_public():
    if User.query.get(get_jwt_identity()) == User.get_by_name('join'):
        return g.dimension.public.save_pkcs1(), 200, {'content-type': 'application/octet-stream'}

    else:
        return {}, 401


servers_to_be_created = {}


def delete_old_temp_servers():
    global servers_to_be_created
    now = get_now()
    for s_id, data in list(servers_to_be_created.items()):
        s_created = dt.datetime.strptime(data['created_on'], defaults.DATETIME_FORMAT)
        if now - s_created > dt.timedelta(hours=1):
            servers_to_be_created.pop(s_id, None)


_lock_delete = threading.Lock()

fetched_catalog = (get_now(), None)
_lock = threading.Lock()


@api_bp.route('/join', methods=['POST'])
@securizer
@jwt_required()
def join():
    global fetched_catalog
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000004':
        js = request.get_json()
        current_app.logger.debug(f"New server wanting to join: {json.dumps(js, indent=2)}")
        if db.session.query(Server).filter_by(id=js.get('id', None)).count() > 0:
            raise errors.DuplicatedId(js.get('id', None))
        if db.session.query(Server).filter_by(name=js.get('name', None)).count() > 0:
            raise errors.AlreadyExists('name', js.get('name', None))
        s = Server.from_json(js)
        s.created_on = get_now()
        external_ip = ipaddress.ip_address(request.remote_addr)
        if not external_ip.is_loopback and external_ip not in [gate.ip for gate in s.gates]:
            for port in set([gate.port for gate in s.gates]):
                s.add_new_gate(external_ip, port, hidden=True)

        certfile = current_app.dm.config.http_conf.get('certfile', None)
        keyfile = current_app.dm.config.http_conf.get('keyfile', None)

        if keyfile and os.path.exists(keyfile):
            with open(keyfile, 'rb') as fh:
                keyfile_content = fh.read()
        else:
            raise errors.FileNotFound(keyfile)
        if certfile and os.path.exists(certfile):
            with open(certfile, 'rb') as fh:
                certfile_content = fh.read()
        else:
            raise errors.FileNotFound(certfile)

        data = {'keyfile': base64.b64encode(keyfile_content).decode(),
                'certfile': base64.b64encode(certfile_content).decode()}

        data.update(Dimension=g.dimension.to_json())
        data.update(me=str(Server.get_current().id))

        with _lock:
            if fetched_catalog[1] is None or fetched_catalog[0] < get_now() - dt.timedelta(minutes=1):
                c = fetch_catalog()
                fetched_catalog = (get_now(), c)
            else:
                c = fetched_catalog[1]
        data.update(catalog=c)

        if _lock_delete.acquire(False):
            try:
                delete_old_temp_servers()
            finally:
                _lock_delete.release()

        server_data = s.to_json(add_gates=True)
        servers_to_be_created.update({s.id: server_data})
        del s
        return data, 200
    else:
        raise errors.GenericError('Invalid token', status_code=400)


_lock_add_node = threading.Lock()


@api_bp.route('/join/acknowledge/<server_id>', methods=['POST'])
@jwt_required()
@lock_catalog
def join_acknowledge(server_id):
    server_data = servers_to_be_created.get(server_id, None)
    if not server_data:
        raise errors.EntityNotFound('Server', server_id)
    s = Server.from_json(server_data)
    with _lock_add_node:
        db.session.add(s)
        db.session.commit()
        current_app.logger.debug(f"Server join acknowledge {s.name}")
    return {}, 204


@api_bp.route('/manager/server_ignore_lock', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
@validate_schema(manager_server_ignore_lock_post)
def manager_server_ignore_lock():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        ignore = request.get_json()['ignore_on_lock']
        for server_id in request.get_json()['server_ids']:
            server = Server.query.get_or_raise(server_id)
            server.l_ignore_on_lock = ignore
            db.session.commit()

        return {}, 204
    else:
        raise errors.UserForbiddenError


@api_bp.route('/send', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
@validate_schema(send_post)
def send():
    def search_cost(ssa, route_list):
        cost = [route['cost'] for route in route_list if str(ssa.server.id) == route['destination_id']]
        if cost:
            if cost[0] is None:
                cost = 999999
            else:
                cost = cost[0]
        else:
            cost = 999999
        return cost

    # Validate Data
    json_data = request.get_json()

    dest_server = Server.query.get_or_raise(json_data['dest_server_id'])

    if 'software_id' in json_data:
        software = Software.query.get_or_raise(json_data['software_id'])

        ssa = SoftwareServerAssociation.query.filter_by(server=g.server, software=software).one_or_none()
        # if current server does not have the software, forward request to the closest server who has it
        if not ssa:
            resp = ntwrk.get(dest_server, 'api_1_0.routes', timeout=5)
            if resp.code == 200:
                ssas = copy.copy(software.ssas)
                ssas.sort(key=functools.partial(search_cost, route_list=resp.msg['route_list']))
            # unable to get route cost, we take the first option we have
            else:
                ssas = random.shuffle(list(software.ssas))
            if not ssas or len(ssas) == 0:
                raise errors.NoSoftwareServer(software_id=str(software.id))
            server = ssas[0].server  # closest server from dest_server who has the software

            resp = ntwrk.post(server, 'api_1_0.send', json=json_data)
            resp.raise_if_not_ok()
            return resp.msg, resp.code
        else:

            file = os.path.join(ssa.path, software.filename)
            if not os.path.exists(file):
                raise errors.FileNotFound(file)
            size = ssa.software.size
    else:
        file = json_data['file']
        if os.path.exists(file):
            size = os.path.getsize(file)
            checksum = md5(json_data.get('file'))
        else:
            raise errors.FileNotFound(file)

    chunk_size = d.CHUNK_SIZE * 1024 * 1024
    max_senders = min(json_data.get('max_senders', d.MAX_SENDERS), d.MAX_SENDERS)
    chunks = math.ceil(size / chunk_size)

    if 'software_id' in json_data:
        json_msg = dict(software_id=str(software.id), num_chunks=chunks)
        if 'dest_path' in json_data:
            json_msg['dest_path'] = json_data.get('dest_path')
    else:
        json_msg = dict(dest_path=json_data['dest_path'], filename=os.path.basename(json_data.get('file')), size=size,
                        checksum=checksum, num_chunks=chunks)
    # if dest_path not set, file will be sent to

    if 'force' in json_data:
        json_msg['force'] = json_data['force']

    resp = ntwrk.post(dest_server, 'api_1_0.transferlist', json=json_msg)
    resp.raise_if_not_ok()

    transfer_id = resp.msg.get('id')
    current_app.logger.debug(
        f"Transfer {transfer_id} created. Sending {file} to {dest_server}:{json_data.get('dest_path')}.")

    if json_data.get('background', True):
        executor.submit(asyncio.run,
                        async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                        chunk_size=chunk_size, max_senders=max_senders, identity=get_jwt_identity()))
    else:
        asyncio.run(async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                    chunk_size=chunk_size, max_senders=max_senders, identity=get_jwt_identity()))

    if json_data.get('include_transfer_data', False):
        resp = ntwrk.get(dest_server, "api_1_0.transferresource", view_data=dict(transfer_id=transfer_id))
        if resp.code == 200:
            msg = resp.msg
        else:
            resp.raise_if_not_ok()
    else:
        msg = {'transfer_id': transfer_id}
    return msg, 202 if json_data.get('background', True) else 201


@api_bp.route('/software/dimensigon', methods=['GET'])
@forward_or_dispatch()
@jwt_required()
@securizer
def software_dimensigon():
    # sends the last software
    repo = os.path.join(current_app.dm.config.config_dir, defaults.SOFTWARE_REPO, defaults.DIMENSIGON_DIR)
    max_version = None
    max_file = None
    for file in os.listdir():
        if 'dimensigon-' in file:
            m = re.search(r'v?\d+\.\d+[.-][ab]?\d+', file)
            if m and parse_version(m.group()) >= parse_version(max_version or dimensigon.__version__):
                max_version = m.group()
                max_file = file
    if max_file:
        with open(os.path.join(repo, 'dimensigon', max_file), 'rb') as fh:
            return {'filename': max_file, 'version': max_version,
                    'content': base64.b64encode(fh.read()).decode('ascii')}, 200
    else:
        return {}, 204


@api_bp.route('/catalog', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
def catalog_update():
    catalog_ver = Catalog.max_catalog()
    new_catalog_ver = current_app.dm.catalog_manager.force_catalog_update()
    if new_catalog_ver:
        if new_catalog_ver > catalog_ver:
            return {'message': "New catalog found",
                    'catalog_version': new_catalog_ver.strftime(defaults.DATEMARK_FORMAT)}, 200
        else:
            return {'message': "No new catalog found"}, 204
    else:
        return {'message': "Catalog upgrade already running"}, 202


@api_bp.route('/catalog/<string:data_mark>', methods=['GET'])
@forward_or_dispatch()
@jwt_required()
@securizer
def catalog(data_mark):
    # Input Validation
    if data_mark == 'initial':
        data_validated = dt.datetime(dt.MINYEAR, 1, 1)
    else:
        try:
            data_validated = dt.datetime.strptime(data_mark, defaults.DATEMARK_FORMAT)
        except Exception as e:
            return {'error': f'Invalid Data Mark: {e}'}, 400

    return fetch_catalog(data_validated)


@log_time()
def fetch_catalog(data_mark=None):
    data = {}
    now = get_now()
    for name, obj in get_distributed_entities():
        c = Catalog.query.get(name)
        # db.session.query to bypass deleted objects to spread deleted changes
        if data_mark:
            query = obj.query.filter(obj.last_modified_at > data_mark)
        else:
            query = obj.query
        repo_data = query.filter(obj.last_modified_at <= now).all()
        if name == 'User':
            data.update({name: [e.to_json(password=True) for e in repo_data]})
        else:
            data.update({name: [e.to_json() for e in repo_data]})
    return data


_cluster_logger = logging.getLogger('dm.cluster')


@api_bp.route('/cluster', methods=['POST'])
# @log_time('full')
# @forward_or_dispatch()
@jwt_required()
# @securizer
# @log_time('after securizer')
# @validate_schema(cluster_post)
# @log_time('after validation')
def cluster():
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        data = request.get_json()
        converted = []
        for item in data:
            try:
                keepalive = dt.datetime.strptime(item['keepalive'], defaults.DATEMARK_FORMAT)
            except ValueError:
                raise errors.InvalidDateFormat(item['keepalive'], defaults.DATEMARK_FORMAT)
            converted.append((item['id'], keepalive, item['death']))
        current_app.dm.cluster_manager.put_many(converted)
        return {}, 204
    else:
        raise errors.UserForbiddenError


@api_bp.route('/cluster/in/<server_id>', methods=['POST'])
@jwt_required()
@securizer
def cluster_in(server_id):
    user = User.get_current()
    data = request.get_json()
    if user and user.name == 'root':
        try:
            keepalive = dt.datetime.strptime(data.get('keepalive'), defaults.DATEMARK_FORMAT)
        except ValueError:
            raise errors.InvalidDateFormat(data.get('keepalive'), defaults.DATEMARK_FORMAT)
        current_app.dm.cluster_manager.put(server_id, keepalive)
        _cluster_logger.debug(
            f"{getattr(Server.query.get(server_id), 'name', server_id) or server_id} is a new alive server")

        current_app.dm.route_manager.new_node_in_cluster(server_id, data['routes'])

        return {'cluster': current_app.dm.cluster_manager.get_cluster(defaults.DATEMARK_FORMAT),
                'neighbours': [s.id for s in Server.get_neighbours()]}, 200

    else:
        raise errors.UserForbiddenError


@api_bp.route('/cluster/out/<server_id>', methods=['POST'])
@jwt_required()
@securizer
def cluster_out(server_id):
    user = User.get_current()
    if user and user.name == 'root':
        Server.query.get_or_raise(server_id)
        data = request.get_json()
        if data.get('death', None):
            try:
                death = dt.datetime.strptime(data['death'], defaults.DATEMARK_FORMAT)
            except:
                death = get_now()
        else:
            death = None
        current_app.dm.cluster_manager.put(server_id, keepalive=death, death=True)

        # run background route
        current_app.dm.route_manager.remove_node_from_cluster(server_id)

        return {}, 202
    else:
        raise errors.UserForbiddenError


# @api_bp.route('/routes/<server_id>', methods=['GET'])
# @forward_or_dispatch()
# @jwt_required()
# @securizer
# def routes_neighbour(server_id):
#     server = Server.query.get_or_raise(server_id)
#     route = routing.check_neighbour(server)
#     # change server
#     if route:
#         return {'neighbour': True}
#     else:
#         return {'neighbour': False}


# set node to alive and check if neighbour

@api_bp.route('/routes', methods=['GET', 'POST', 'PATCH'])
# @log_time('full')
@forward_or_dispatch('GET', 'POST')
@jwt_required()
@securizer
@validate_schema(POST=routes_post, PATCH=routes_patch)
# @log_time('after validation')
def routes():
    if request.method == 'GET':
        route_table = []
        for route in Route.query.join(Server.route).order_by(Server.name).filter(Server.deleted == False).all():
            route_table.append(route.to_json(human=check_param_in_uri('human')))
        data = {'route_list': route_table}
        data.update(server=dict(name=g.server.name, id=str(g.server.id)))
        return data

    elif request.method == 'POST':
        data = request.get_json()
        current_app.dm.route_manager.refresh_table(discover_new_neighbours=data.get('discover_new_neighbours', False),
                                                   check_current_neighbours=data.get('check_current_neighbours', False),
                                                   max_num_discovery=data.get('max_num_discovery', 5))
        return {}, 204

    elif request.method == 'PATCH':
        current_app.dm.route_manager.new_routes(request.get_json())
        return {}, 204


def run_command_and_callback(operation: 'IOperationEncapsulation', params, context: Context, source: Server,
                             step_execution: StepExecution,
                             event_id,
                             identity,
                             timeout=None):
    execution: StepExecution = db.session.merge(step_execution)
    exec_id = execution.id
    source = db.session.merge(source)
    start = get_now()
    try:
        cp = operation.execute(params, timeout=timeout, context=context)
    except Exception as e:
        cp = CompletedProcess(success=False, stderr=f"Error while executing operation. {format_exception(e)}",
                              start_time=start, end_time=get_now())
    finally:
        execution.load_completed_result(cp)

    data = dict(step_execution=execution.to_json())
    if execution.child_orch_execution:
        data['step_execution'].update(orch_execution=execution.child_orch_execution.to_json(add_step_exec=True))

    # commit after data is dumped
    try:
        db.session.commit()
    except Exception as e:
        current_app.logger.exception(f"Error on commit for execution {exec_id}")

    resp, code = ntwrk.post(server=source, view_or_url='api_1_0.events', view_data={'event_id': event_id}, json=data,
                            identity=identity)
    if code != 202:
        current_app.logger.error(f"Error while sending result for execution {exec_id}: {code}, {resp}")
    return data


@api_bp.route('/launch/operation', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
def launch_operation():
    data = request.get_json()
    operation = pickle.loads(base64.b64decode(data['operation'].encode('ascii')))
    var_context = pickle.loads(base64.b64decode(data['var_context'].encode('ascii')))
    params = pickle.loads(base64.b64decode(data['params'].encode('ascii')))
    orch_exec_json = data['orch_execution']
    orch_exec = OrchExecution.from_json(orch_exec_json)

    se = StepExecution(id=var_context.env.get('step_execution_id'), server=g.server, step_id=data.get('step_id'),
                       orch_execution=orch_exec, params=params,
                       start_time=get_now())
    db.session.add_all([orch_exec, se])
    db.session.commit()
    future = executor.submit(run_command_and_callback, operation, params, var_context, g.source, se,
                             data['event_id'], identity=get_jwt_identity(),
                             timeout=data.get('timeout', None))
    try:
        r = future.result(1)
    except concurrent.futures.TimeoutError:
        return {}, 204
    except Exception as e:
        current_app.logger.exception(
            f"Exception got when executing step {data.get('step_id')}. See logs for more information")
        return {'error': f"Exception got when executing step {data.get('step_id')}. See logs for more information"}, 500
    else:
        return r, 200


@api_bp.route('/launch/orchestration', defaults={'orchestration_id': None}, methods=['POST'])
@api_bp.route('/launch/orchestration/<orchestration_id>', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
@validate_schema(launch_orchestration_post)
def launch_orchestration(orchestration_id):
    data = request.get_json()
    if orchestration_id:
        orchestration = Orchestration.query.get_or_raise(orchestration_id)
    else:
        iden = (data.get('orchestration'),)
        columns = ('orchestration',)
        query = Orchestration.query.filter_by(name=data.get('orchestration'))
        if 'version' in data:
            iden += (data.get('version'),)
            columns += ('version',)
            query = query.filter_by(version=data.get('version'))
        query = query.order_by(Orchestration.version.desc())
        if query.count() <= 1:
            orchestration = query.one_or_none()
        else:
            orchestration = query.first()
        if not orchestration:
            raise errors.EntityNotFound('Orchestration', iden, columns)

    if not orchestration.steps:
        return errors.GenericError('orchestration does not have steps to execute', orchestration_id=orchestration_id)

    params = data.get('params') or {}
    hosts = data.get('hosts', Server.get_current().id)

    a = set(orchestration.target)
    if not isinstance(hosts, dict):
        hosts = dict(all=hosts)
    b = set(hosts.keys())
    c = a - b
    if len(c) > 0:
        raise errors.TargetUnspecified(c)
    c = b - a
    if len(c) > 0:
        raise errors.TargetNotNeeded(c)

    not_found = normalize_hosts(hosts)
    if not_found:
        raise errors.ServerNormalizationError(not_found)

    for target, target_hosts in hosts.items():
        if len(target_hosts) == 0:
            raise errors.EmptyTarget(target)
    # check param entries
    # rest = orchestration.user_parameters - set(params.keys())
    # if rest:
    #     rest = list(rest)
    #     rest.sort()
    #     return {'error': f"Parameter(s) not specified: {', '.join(rest)}"}, 404

    execution_id = str(uuid.uuid4())

    executor_id = get_jwt_identity()
    vc = Context(params, dict(execution_id=None, root_orch_execution_id=execution_id, orch_execution_id=execution_id,
                              executor_id=executor_id),
                 vault=Vault.get_variables_from(executor_id, scope=data.get('scope', 'global')))

    if not data.get('skip_validation', False):
        validate_input_chain(orchestration,
                             {'input': set(params.keys()), 'env': set(vc.env.keys()), 'vault': set(vc.vault.keys())})

    if request.get_json().get('background', True):
        future = executor.submit(deploy_orchestration, orchestration=orchestration.id, var_context=vc, hosts=hosts,
                                 timeout=data.get('timeout', None))
        try:
            future.result(5)
        except concurrent.futures.TimeoutError:
            return {'execution_id': execution_id}, 202
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
    else:
        try:
            deploy_orchestration(orchestration=orchestration, var_context=vc, hosts=hosts,
                                 timeout=data.get('timeout', None))
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
    return OrchExecution.query.get(execution_id).to_json(add_step_exec=True, human=check_param_in_uri('human'),
                                                         split_lines=True), 200


def wrap_sudo(user, cmd):
    if isinstance(user, User):
        username = user.name
    else:
        username = user
    if isinstance(cmd, str):
        return f"sudo -Siu {username} -- sh -c {shlex.quote(cmd)}"
    else:
        args = ["sudo", "-Siu", username, "--", "bash" "-c", ] + cmd
        return args


@api_bp.route('/launch/command', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
@validate_schema(launch_command_post)
def launch_command():
    data = request.get_json()

    server_list = []
    if 'target' in data:
        not_found = []
        servers = Server.query.all()
        if data['target'] == 'all':
            server_list = servers
        elif is_iterable_not_string(data['target']):
            for vv in data['target']:
                sl = search(vv, servers)
                if len(sl) == 0:
                    not_found.append(vv)
                else:
                    server_list.extend(sl)
        else:
            sl = search(data['target'], servers)
            if len(sl) == 0:
                not_found.append(data['target'])
            else:
                server_list.extend(sl)
        if not_found:
            return {'error': "Following granules or ids did not match to any server: " + ', '.join(not_found)}, 404
    else:
        server_list.append(g.server)

    if re.search(r'rm\s+((-\w+|--[-=\w]*)\s+)*(-\w*[rR]\w*|--recursive)', data['command']):
        return {'error': 'rm with recursion is not allowed'}, 403
    data.pop('target', None)
    start = None

    username = getattr(User.query.get(get_jwt_identity()), 'name', None)
    if not username:
        raise errors.EntityNotFound('User', get_jwt_identity())
    cmd = wrap_sudo(username, data['command'])
    if g.server in server_list:
        start = time.time()
        server_list.pop(server_list.index(g.server))
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE if data.get('input', None) else None, stderr=subprocess.PIPE,
                                stdout=subprocess.PIPE, shell=True, close_fds=True, encoding='utf-8')

    resp_data = {}
    if check_param_in_uri("human"):
        attr = 'name'
    else:
        attr = 'id'
    if server_list:
        resp: t.List[ntwrk.Response] = asyncio.run(
            ntwrk.parallel_requests(server_list, method='POST', view_or_url='api_1_0.launch_command', json=data))
        for s, r in zip(server_list, resp):
            key = getattr(s, attr, s.id)
            if r.ok:
                resp_data[key] = r.msg[s.id]
            else:
                if not r.exception:
                    resp_data[key] = {
                        'error': {'status_code': r.code, 'response': r.msg}}
                else:
                    if isinstance(r.exception, errors.BaseError):
                        resp_data[key] = errors.format_error_content(r.exception, current_app.config['DEBUG'])
                    else:
                        resp_data[key] = {
                            'error': format_exception(r.exception) if current_app.config['DEBUG'] else str(
                                r.exception) or str(r.exception.__class__.__name__)}

    if start:
        key = getattr(g.server, attr, g.server.id)
        timeout = data.get('timeout', defaults.TIMEOUT_COMMAND)
        try:
            outs, errs = proc.communicate(input=(data.get('input', '') or ''), timeout=timeout)
        except (TimeoutError, subprocess.TimeoutExpired):
            proc.kill()
            try:
                outs, errs = proc.communicate(timeout=1)
            except:
                resp_data[key] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds. Unable to communicate with the process launched."}
            else:
                resp_data[key] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds",
                    'stdout': outs.split('\n'), 'stderr': errs.split('\n')}
        except Exception as e:
            current_app.logger.exception("Exception raised while trying to run command")
            resp_data[key] = {
                'error': traceback.format_exc() if current_app.config['DEBUG'] else str(
                    r.exception) or str(r.exception.__class__.__name__)}
        else:
            resp_data[key] = {'stdout': outs.split('\n'), 'stderr': errs.split('\n'),
                              'returncode': proc.returncode}
    resp_data['cmd'] = cmd
    resp_data['input'] = data.get('input', None)
    return resp_data, 200


@api_bp.route('/events/<event_id>', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
def events(event_id):
    e = Event(event_id, data=request.get_json())
    current_app.events.dispatch(e)
    return {}, 202


@api_bp.route('/orchestrations/full', methods=['POST'])
@forward_or_dispatch()
@jwt_required()
@securizer
@validate_schema(orchestration_full)
@lock_catalog
def orchestrations_full():
    json_data = request.get_json()
    json_steps = json_data.pop('steps')
    generated_version = False
    if 'version' not in json_data:
        generated_version = True
        json_data['version'] = Orchestration.query.filter_by(name=json_data['name']).count() + 1
    o = Orchestration(**json_data)
    db.session.add(o)
    resp_data = {'id': o.id}
    if generated_version:
        resp_data.update(version=o.version)

    # reorder steps in order of dependency
    id2step = {s['id']: s for s in json_steps}

    dag = DAG()
    for s in json_steps:
        step_id = s['id']
        if s['undo'] and len(s.get('parent_step_ids', [])) == 0:
            raise errors.UndoStepWithoutParent(step_id)
        dag.add_node(step_id)
        for p_s_id in s.get('parent_step_ids', []):
            dag.add_edge(p_s_id, step_id)

    if dag.is_cyclic():
        raise errors.CycleError

    new_steps = []
    for step_id in dag.ordered_nodes:
        step = id2step[step_id]
        new_steps.append(step)
    # end reorder steps in order of dependency

    rid2step = OrderedDict()
    dependencies = {}
    for json_step in new_steps:
        rid = json_step.pop('id', None)
        json_step['rid'] = rid
        if rid is not None and rid in rid2step.keys():
            raise errors.DuplicatedId(rid)
        if 'action_template_id' in json_step:
            json_step['action_template'] = ActionTemplate.query.get_or_raise(json_step.pop('action_template_id'))
        elif 'action_type' in json_step:
            json_step['action_type'] = ActionType[json_step.pop('action_type')]
        dependencies[rid] = {'parent_step_ids': [p_id for p_id in json_step.pop('parent_step_ids', [])]}
        s = o.add_step(**json_step)
        db.session.add(s)
        if rid:
            rid2step[rid] = s

        continue

    # process dependencies
    for rid, dep in dependencies.items():
        step = rid2step[rid]
        parents = []
        for p_s_id in dep['parent_step_ids']:
            if p_s_id in rid2step:
                parents.append(rid2step[p_s_id])
        o.set_parents(step, parents)

    db.session.commit()

    # send new ids in order of appearance at beginning
    new_id_steps = []
    for rid in rid2step.keys():
        new_id_steps.append(rid2step[rid].id)
    resp_data.update({'step_ids': new_id_steps})
    return resp_data, 201



import base64
import logging
import os
import zlib

from flask import request
from flask_jwt_extended import jwt_required, get_jwt_identity
from flask_restful import Resource

from dimensigon.domain.entities import Server, File, FileServerAssociation
from dimensigon.web import db, errors
from dimensigon.web.api_1_0 import api_bp
from dimensigon.web.decorators import forward_or_dispatch, securizer, validate_schema, lock_catalog
from dimensigon.web.helpers import filter_query, check_param_in_uri
from dimensigon.web.json_schemas import files_post, file_post, file_patch, file_sync

_logger = logging.getLogger('dm.fileSync')


@api_bp.route("/file/<file_id>/sync", methods=['POST'])
@jwt_required()
@securizer
@forward_or_dispatch()
@validate_schema(file_sync)
def file_sync(file_id):
    if get_jwt_identity() == '00000000-0000-0000-0000-000000000001':
        data = request.get_json()
        file = File.query.get(file_id)
        if file is None and not data.get('force', False):
            raise errors.EntityNotFound("File", file_id)

        file = data.get('file')
        content = zlib.decompress(base64.b64decode(data.get('data').encode('ascii')))

        _logger.debug(f"received file sync {file}.")
        try:
            if not os.path.exists(os.path.dirname(file)):
                os.makedirs(os.path.dirname(file))
            with open(file, 'wb') as fh:
                fh.write(content)
        except Exception as e:
            raise errors.GenericError(f"Error while trying to create/write file: {e}", 500)
        return {}, 204
    else:
        raise errors.UserForbiddenError


class FileList(Resource):

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    def get(self):
        query = filter_query(File, request.args)
        return [file.to_json(human=check_param_in_uri('human'), no_delete=True,
                             destinations=check_param_in_uri('destinations')) for file in
                query.all()]

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(files_post)
    @lock_catalog
    def post(self):
        data = request.get_json()
        source_server = Server.query.get_or_raise(data.pop('src_server_id'))
        entities = [File(source_server=source_server, target=data['target'], dest_folder=data.get('dest_folder', None))]
        for dest in data.get('destinations', []):
            dest_server = Server.query.get_or_raise(dest['dst_server_id'])
            entities.append(FileServerAssociation(file=entities[0], destination_server=dest_server,
                                                  dest_folder=dest.get('dest_folder', None)))
        db.session.add_all(entities)
        db.session.commit()
        return {'id': str(entities[0].id)}, 201


class FileResource(Resource):

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    def get(self, file_id):
        return File.query.get_or_raise(file_id).to_json(human=check_param_in_uri('human'), delete_data=False,
                                                        destinations=True)

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @validate_schema(file_post)
    def post(self, file_id):
        from .file_server_association import change_destinations
        file = File.query.get_or_raise(file_id)
        data = request.get_json()
        file.source_server = Server.query.get_or_raise(data.get('src_server_id'))
        if not data['target']:
            raise errors.InvalidValue("The field cannot be empty", field='target', value=data['target'])
        file.target = data.get('target')
        file.dest_folder = data.get('dest_folder')
        change_destinations(file, data.get('destinations', []))
        if file in db.session.dirty:
            db.session.commit()
            return {}, 204
        return {}, 202

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @validate_schema(file_patch)
    @lock_catalog
    def patch(self, file_id):
        from .file_server_association import change_destinations
        file = File.query.get_or_raise(file_id)
        data = request.get_json()
        if 'src_server_id' in data and file.source_server.id != data['src_server_id']:
            s = Server.query.get_or_raise(data['src_server_id'])
            file.source_server = s
        if 'target' in data and file.target != data['target']:
            if not data['target']:
                raise errors.InvalidValue("The field cannot be empty", field='target', value=data['target'])
            file.target = data['target']
        if 'dest_folder' in data and file.dest_folder != data.get('dest_folder'):
            file.dest_folder = data.get('dest_folder')
        if 'destinations' in data:
            change_destinations(file, data['destinations'])
        if file in db.session.dirty:
            db.session.commit()
            return {}, 204
        return {}, 202

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @lock_catalog
    def delete(self, file_id):
        file = File.query.get_or_raise(file_id)
        file.delete()
        db.session.commit()
        return {}, 204



import abc
import functools
import inspect
import logging
import multiprocessing as mp
import multiprocessing.queues as mpq
import signal
import sys
import threading
import time
import typing as t
from queue import Empty, Full

from dimensigon.use_cases import mptools_events as events
from dimensigon.utils.helpers import is_iterable_not_string

DEFAULT_POLLING_TIMEOUT = 0.1
MAX_SLEEP_SECS = 0.1

_logger = logging.getLogger('dm.mptools')


class MPQueue(mpq.Queue):

    # -- See StackOverflow Article :
    #   https://stackoverflow.com/questions/39496554/cannot-subclass-multiprocessing-queue-in-python-3-5
    #
    # -- tldr; mp.Queue is a _method_ that returns an mpq.Queue object.  That object
    # requires a context for proper operation, so this __init__ does that work as well.
    def __init__(self, *args, **kwargs):
        ctx = mp.get_context()
        super().__init__(*args, **kwargs, ctx=ctx)

    def safe_get(self, timeout=DEFAULT_POLLING_TIMEOUT):
        try:
            if timeout is None:
                return self.get(block=False)
            else:
                return self.get(block=True, timeout=timeout)
        except Empty:
            return None

    def safe_put(self, item, timeout=DEFAULT_POLLING_TIMEOUT):
        try:
            self.put(item, block=True, timeout=timeout)
            return True
        except Full:
            return False

    def drain(self):
        item = self.safe_get()
        while item:
            yield item
            item = self.safe_get()

    def safe_close(self):
        num_left = sum(1 for __ in self.drain())
        self.close()
        self.join_thread()
        return num_left


# -- useful function
def _sleep_secs(max_sleep, end_time=999999999999999.9):
    # Calculate time left to sleep, no less than 0
    return max(0.0, min(end_time - time.time(), max_sleep))


# -- Signal Handling
class TerminateInterrupt(BaseException):
    pass


class SignalObject:
    MAX_TERMINATE_CALLED = 3

    def __init__(self, shutdown_event):
        self.terminate_called = 0
        self.shutdown_event = shutdown_event


def default_signal_handler(signal_object, exception_class, signal_num, current_stack_frame):
    signal_object.terminate_called += 1
    signal_object.shutdown_event.set()
    # if signal_object.loop:
    #     signal_object.loop.stop()
    _logger.info("shutdown event set")
    if signal_object.terminate_called >= signal_object.MAX_TERMINATE_CALLED:
        raise exception_class()


def init_signal(signal_num, signal_object, exception_class, handler):
    handler = functools.partial(handler, signal_object, exception_class)
    signal.signal(signal_num, handler)
    signal.siginterrupt(signal_num, False)


def init_signals(shutdown_event, int_handler, term_handler):
    signal_object = SignalObject(shutdown_event)
    init_signal(signal.SIGINT, signal_object, KeyboardInterrupt, int_handler)
    init_signal(signal.SIGTERM, signal_object, TerminateInterrupt, term_handler)
    return signal_object


class Observer(list):

    def __call__(self, event: events.EventMessage):
        for item in self:
            item(event)


EventType = t.Union[str, t.Type[events.EventMessage], events.EventMessage]


class EventHandler(threading.Thread):

    def __init__(self, queue: MPQueue, *args, **kwargs):
        super().__init__()
        self.queue = queue
        self.stop_event = threading.Event()
        self._event_handlers = {}

    @staticmethod
    def _e_type(event):
        if inspect.isclass(event) and issubclass(event, events.EventMessage):
            e_type = event.__name__
        elif isinstance(event, events.EventMessage):
            e_type = event.event_type
        else:
            e_type = event
        return e_type

    def listen(self, event_type: t.Union[t.Iterable[EventType], EventType],
               callback):
        if not is_iterable_not_string(event_type):
            event_type = [event_type]
        for et in event_type:
            sig = inspect.signature(callback)
            assert len(sig.parameters) == 1, "Callback must get 1 positional argument"
            e_type = self._e_type(et)

            if e_type not in self._event_handlers:
                self._event_handlers[e_type] = Observer()
            self._event_handlers[e_type].append(callback)

    def detach(self, event_type: t.Union[t.Iterable[EventType], EventType],
               callback):
        if not is_iterable_not_string(event_type):
            event_type = [event_type]
        for et in event_type:
            e_type = self._e_type(et)
            self._event_handlers.get(e_type, []).remove(callback)

    def run(self):
        while not self.stop_event.is_set():
            event = self.queue.safe_get()
            if event:
                # _logger.debug(f"Processing event {event}")
                e_type = self._e_type(event)
                if event in (events.StopEventHandler, events.Stop):
                    break
                else:
                    [h(event) for h in self._event_handlers.get(e_type, [])]

    def stop(self):
        self.stop_event.set()
        self.queue.safe_put(events.StopEventHandler())


# -- Worker classes

class Worker:
    MAX_TERMINATE_CALLED = 3
    int_handler = staticmethod(default_signal_handler)
    term_handler = staticmethod(default_signal_handler)

    def __init__(self, name, startup_event: mp.Event, shutdown_event: mp.Event, publish_q: MPQueue, event_q: MPQueue,
                 *args, **kwargs):
        self.name = name
        self.logger = logging.getLogger(f'dm.{self.name}')
        self.startup_event = startup_event
        self.shutdown_event = shutdown_event
        self.publish_q = publish_q
        self.event_q = event_q
        self.dispatcher = EventHandler(self.event_q)
        self.terminate_called = 0
        self.init_args(*args, **kwargs)

    def _init_signals(self):
        self.logger.debug("Entering init_signals")
        try:
            signal_object = init_signals(self.shutdown_event, self.int_handler, self.term_handler)
        except ValueError:
            pass
        else:
            return signal_object
        # loop = asyncio.get_event_loop()
        # for signame in {'SIGINT', 'SIGTERM'}:
        #     loop.add_signal_handler(
        #         getattr(signal, signame),
        #         functools.partial(default_async_signal_handler, signame, loop))

    def init_args(self, *args, **kwargs):
        pass

    def startup(self):
        self.logger.debug("Entering startup")
        pass

    def _startup(self):
        self.dispatcher.start()
        self.startup()

    def shutdown(self):
        self.logger.debug("Entering shutdown")
        pass

    def _shutdown(self):
        self.shutdown()
        self.dispatcher.stop()
        self.dispatcher.join()

    def _main_loop(self):
        self.logger.debug("Entering main_loop")
        while not self.shutdown_event.is_set():
            self.main_func()

    @abc.abstractmethod
    def main_func(self, *args, **kwargs):
        self.logger.debug("Entering main_func")
        raise NotImplementedError(f"{self.__class__.__name__}.main_func is not implemented")

    def run(self):
        self._init_signals()
        try:
            self._startup()
            self.startup_event.set()
            self._main_loop()
            self.logger.info("Normal Shutdown")
            self.publish_q.safe_put(events.EventMessage("SHUTDOWN", msg_src=self.name, msg="Normal"))
            return 0
        except BaseException as exc:
            # -- Catch ALL exceptions, even Terminate and Keyboard interrupt
            self.logger.error(f"Exception Shutdown: {exc}", exc_info=True)
            self.publish_q.safe_put(events.EventMessage("FATAL", msg_src=self.name, msg=exc))
            if type(exc) in (TerminateInterrupt, KeyboardInterrupt):
                sys.exit(1)
            else:
                sys.exit(2)
        finally:
            self._shutdown()


class TimerWorker(Worker):
    INTERVAL_SECS = 10
    MAX_SLEEP_SECS = 0.02

    def _main_loop(self):
        self.next_time = time.time() + self.INTERVAL_SECS
        while not self.shutdown_event.is_set():
            sleep_secs = _sleep_secs(self.MAX_SLEEP_SECS, self.next_time)
            time.sleep(sleep_secs)
            if self.next_time and time.time() > self.next_time:
                self.logger.log(1, f"Calling main_func")
                self.main_func()
                self.next_time = time.time() + self.INTERVAL_SECS


# class QueueWorker(Worker):
#     def init_args(self, args):
#         self.logger.debug(f"Entering QueueProcWorker.init_args : {args}")
#         self.work_q, = args
#
#     def _main_loop(self):
#         self.logger.debug("Entering QueueProcWorker.main_loop")
#         while not self.shutdown_event.is_set():
#             item = self.work_q.safe_get()
#             if not item:
#                 continue
#             self.logger.debug(f"QueueProcWorker.main_loop received '{item}' message")
#             if item == "END":
#                 break
#             else:
#                 self.main_func(item)


# -- Process Wrapper

def proc_worker_wrapper(proc_worker_class, name, startup_evt, shutdown_evt, publish_q, event_q, *args, **kwargs):
    proc_worker = proc_worker_class(name, startup_evt, shutdown_evt, publish_q, event_q, *args, **kwargs)
    return proc_worker.run()


class Proc:
    STARTUP_WAIT_SECS = 3
    SHUTDOWN_WAIT_SECS = 90

    def __init__(self, name: str, worker_class: t.Type[Worker], shutdown_event: mp.Event, publish_q: MPQueue,
                 event_q: MPQueue, async_loop=False, *args, **kwargs):
        self.name = name
        self.logger = logging.getLogger(f'dm.{self.name}')
        self.shutdown_event = shutdown_event
        self.startup_event = mp.Event()
        self._proc_worker = worker_class(f"{name}", self.startup_event, shutdown_event, publish_q, event_q, async_loop,
                                         *args, **kwargs)
        self.proc = mp.Process(target=self._proc_worker.run, name=name)
        self.logger.debug(f"Starting {name} process")
        self.proc.start()
        started = self.startup_event.wait(timeout=Proc.STARTUP_WAIT_SECS)
        self.logger.debug(f"{name} {'started' if started else 'NOT started'}")
        if not started:
            self.terminate()
            raise RuntimeError(f"Process {name} failed to startup after {Proc.STARTUP_WAIT_SECS} seconds")

    def __getattr__(self, item):
        return getattr(self._proc_worker, item)

    def full_stop(self, wait_time=SHUTDOWN_WAIT_SECS):
        self.logger.debug(f"Stopping process {self.name}")
        self.shutdown_event.set()
        self.proc.join(wait_time)
        if self.proc.is_alive():
            self.terminate()

    def terminate(self):
        self.logger.debug(f"Terminating process {self.name}")
        NUM_TRIES = 3
        tries = NUM_TRIES
        while tries and self.proc.is_alive():
            self.proc.terminate()
            time.sleep(0.01)
            tries -= 1

        if self.proc.is_alive():
            self.logger.error(f"Failed to terminate {self.name} after {NUM_TRIES} attempts")
            return False
        else:
            self.logger.info(f"Terminated {self.name} after {NUM_TRIES - tries} attempt(s)")
            return True


class Thread(Proc):
    SHUTDOWN_WAIT_SECS = 90

    def __init__(self, name: str, worker_class: t.Type[Worker], shutdown_event: mp.Event, publish_q: MPQueue,
                 event_q: MPQueue, async_loop=False, *args, **kwargs):
        self.name = name
        self.logger = logging.getLogger(f'dm.{self.name}')
        self.shutdown_event = shutdown_event
        self.startup_event = mp.Event()
        self._proc_worker = worker_class(f"{name}", self.startup_event, shutdown_event, publish_q, event_q,
                                         async_loop,
                                         *args, **kwargs)
        self.proc = threading.Thread(target=self._proc_worker.run, name=name)
        self.logger.debug(f"Starting {name} thread")
        self.proc.start()
        self.logger.debug(f"{name} started")

    def __getattr__(self, item):
        return getattr(self._proc_worker, item)

    def full_stop(self, wait_time=SHUTDOWN_WAIT_SECS):
        self.logger.debug(f"Stopping thread {self.name}")
        self.shutdown_event.set()
        self.proc.join(wait_time)
        if self.proc.is_alive():
            self.logger.warning(f"Thread {self.name} did not stop")

    def terminate(self):
        pass


# -- Main Wrappers
class MainContext:
    STOP_WAIT_SECS = 3.0

    def __init__(self, logger=None):

        self.logger = logging.getLogger(logger or 'dm.main')
        self.procs: t.List[Proc] = []
        self.threads: t.List[Thread] = []
        self.queues: t.List[MPQueue] = []
        self.shutdown_event = mp.Event()
        self.publish_q = MPQueue()

    def forward_events(self):
        item = True
        while item:
            item = self.publish_q.safe_get()
            if item:
                self.logger.debug(f"Spread event {item}")
                if item == events.Stop:
                    break
                [q.safe_put(item) for q in self.queues]

    def init_signals(self):
        return init_signals(self.shutdown_event, default_signal_handler, default_signal_handler)

    def start(self):
        pass

    def stop(self):
        self._stopped_procs_result = self.stop_procs()
        self._stopped_thread_result = self.stop_threads()
        self._stopped_queues_result = self.stop_queues()

    def __enter__(self):

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.logger.exception(f"Exception: {exc_val}", exc_info=(exc_type, exc_val, exc_tb))

        self.stop()

        # -- Don't eat exceptions that reach here.
        return not exc_type

    def Proc(self, worker_class: t.Type[Worker], *args, **kwargs):
        queue = self.MPQueue()
        if 'name' in 'kwargs':
            name = kwargs.pop('name')
        else:
            name = worker_class.__name__
        proc = Proc(name, worker_class, self.shutdown_event, self.publish_q, queue, *args, **kwargs)
        self.procs.append(proc)
        return proc

    def Thread(self, worker_class: t.Type[Worker], *args, **kwargs):
        queue = self.MPQueue()
        if 'name' in 'kwargs':
            name = kwargs.pop('name')
        else:
            name = worker_class.__name__
        proc = Thread(name, worker_class, self.shutdown_event, self.publish_q, queue, *args, **kwargs)
        self.threads.append(proc)
        return proc

    def MPQueue(self, *args, **kwargs):
        q = MPQueue(*args, **kwargs)
        self.queues.append(q)
        return q

    def publish(self, event):
        [q.safe_put(event) for q in self.queues]

    def stop_procs(self):
        # self.publish(events.Stop(msg_src="stop_procs", msg="END"))
        self.shutdown_event.set()
        end_time = time.time() + self.STOP_WAIT_SECS
        num_terminated = 0
        num_failed = 0

        # -- Wait up to STOP_WAIT_SECS for all processes to complete
        for proc in self.procs:
            join_secs = _sleep_secs(self.STOP_WAIT_SECS, end_time)
            proc.proc.join(join_secs)

        # -- Clear the procs list and _terminate_ any procs that
        # have not yet exited
        still_running = []
        while self.procs:
            proc = self.procs.pop()
            if proc.proc.is_alive():
                if proc.terminate():
                    num_terminated += 1
                else:
                    still_running.append(proc)
            else:
                if hasattr(proc.proc, 'exitcode'):
                    exitcode = proc.proc.exitcode
                    if exitcode:
                        self.logger.error(f"Process {proc.name} ended with exitcode {exitcode}")
                        num_failed += 1
                    else:
                        self.logger.debug(f"Process {proc.name} stopped successfully")

        self.procs = still_running
        return num_failed, num_terminated

    def stop_threads(self):
        # self.publish(events.Stop(msg_src="stop_procs", msg="END"))
        self.shutdown_event.set()
        end_time = time.time() + self.STOP_WAIT_SECS
        num_terminated = 0
        num_failed = 0

        # -- Wait up to STOP_WAIT_SECS for all processes to complete
        for th in self.threads:
            join_secs = _sleep_secs(self.STOP_WAIT_SECS, end_time)
            th.proc.join(join_secs)

        # -- Clear the procs list and _terminate_ any procs that
        # have not yet exited
        still_running = []
        while self.threads:
            th = self.threads.pop()
            if th.proc.is_alive():
                still_running.append(th)
            else:
                num_terminated += 1

        self.threads = still_running
        return num_terminated

    def stop_queues(self):
        num_items_left = 0
        # -- Clear the queues list and close all associated queues
        for q in self.queues:
            num_items_left += sum(1 for __ in q.drain())
            q.close()

        # -- Wait for all queue threads to stop
        while self.queues:
            q = self.queues.pop(0)
            q.join_thread()
        return num_items_left



import typing as t

from flask import request, current_app
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dimensigon.domain.entities import FileServerAssociation, File, Server
from dimensigon.web import db, errors
from dimensigon.web.decorators import securizer, forward_or_dispatch, lock_catalog, validate_schema
from dimensigon.web.helpers import filter_query, check_param_in_uri
from dimensigon.web.json_schemas import file_server_associations_post, file_server_associations_patch, \
    file_server_associations_delete


class FileServerAssociationList(Resource):

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    def get(self, file_id):
        query = filter_query(FileServerAssociation, request.args).filter_by(file_id=file_id)
        return [fsa.to_json(human=check_param_in_uri('human'), no_delete=True) for fsa in query.all()]

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @validate_schema(file_server_associations_post)
    @lock_catalog
    def post(self, file_id):
        destinations = request.get_json()
        if isinstance(destinations, dict):
            destinations = [destinations]
        new_fsas = change_destinations(File.query.get_or_raise(file_id), destinations, action='add')
        new = [(fsa.file.id, fsa.destination_server.id) for fsa in new_fsas]
        db.session.commit()
        fs = current_app.dm.file_sync
        if fs:
            [fs.add(*k) for k in new]
        return {}, 204

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @validate_schema(file_server_associations_patch)
    @lock_catalog
    def patch(self, file_id):
        destinations = request.get_json()
        if isinstance(destinations, dict):
            destinations = [destinations]
        new_fsas = change_destinations(File.query.get_or_raise(file_id), destinations)
        new = [(fsa.file.id, fsa.destination_server.id) for fsa in new_fsas]
        db.session.commit()
        fs = current_app.extensions.get('file_sync')
        if fs:
            [fs.add(*k) for k in new]
        return {}, 204

    @jwt_required()
    @securizer
    @forward_or_dispatch()
    @validate_schema(file_server_associations_delete)
    @lock_catalog
    def delete(self, file_id):
        destinations = request.get_json()
        if isinstance(destinations, dict):
            destinations = [destinations]
        change_destinations(File.query.get_or_raise(file_id), destinations, action='delete')
        db.session.commit()
        return {}, 204


def change_destinations(file: File, destinations: t.List, action: str = None):
    """

    :param file: file to change its destinations
    :param destinations:
    :param action:
    :return:
    """
    current = set([d.dst_server_id for d in file.destinations])
    new = set([d.get('dst_server_id') for d in destinations])
    new_fsas = []
    if action == 'add':
        to_remove = []
        to_modify = []
        already_there = current.intersection(new)
        if already_there:
            raise errors.InvalidValue("destination servers already exist",
                                      destinations=[{'id': ident, 'name': Server.query.get(ident).name} for ident
                                                    in already_there])
        to_add = new
    elif action == 'delete':
        to_remove = new
        not_there = new - current
        if not_there:
            raise errors.InvalidValue("destination servers do not exist",
                                      destinations=[{'id': ident, 'name': Server.query.get(ident).name} for ident
                                                    in not_there])
        to_modify = []
        to_add = []
    else:
        to_remove = current - new
        to_modify = new.intersection(current)
        to_add = new - current

    for dst_server_id in to_add:
        s = Server.query.get_or_raise(dst_server_id)
        if s._me:
            raise errors.InvalidValue("Destination cannot be the same as the source server",
                                      source_server={'id': s.id, 'name': s.name})
        dest = [d for d in destinations if d['dst_server_id'] == dst_server_id][0]
        fsa = FileServerAssociation(file=file, destination_server=s, dest_folder=dest.get('dest_folder', None))

        db.session.add(fsa)
        new_fsas.append(fsa)
    for dst_server_id in to_modify:
        s = Server.query.get_or_raise(dst_server_id)
        curr_dest = [d for d in file.destinations if d.dst_server_id == dst_server_id][0]
        in_dest = [d for d in destinations if d['dst_server_id'] == dst_server_id][0]
        if curr_dest.dest_folder != in_dest.get('dest_folder', None):
            curr_dest.dest_folder = in_dest.get('dest_folder', None)
    for dst_server_id in to_remove:
        s = Server.query.get_or_raise(dst_server_id)
        dest = [d for d in file.destinations if d.dst_server_id == dst_server_id]
        if dest:
            db.session.delete(dest[0])
        else:
            raise errors.GenericError("File does not have this destination", status_code=404, file_id=file.id,
                                      destination={'id': dst_server_id, 'name': s.name})
    return new_fsas



"""
Usage: dshell orch list [--id ID|--name NAME|--like LIKE] [--version N] [--detail]
       dshell orch create (- | FILE)
       dshell orch run ID [[--param PARAM=VALUE]... | [--json-parameters JSON]]
                          ((--target NAME=VALUE)... | --json-target JSON)
                          [--no-wait]

Command to interact with orchestration

Arguments:
    FILE                     File which has the orchestration to create
    ID                       Orchestration ID to launch
    NAME                     Orchestration's name

Options:
    --id ID                  Action template id to show
    --last N                 Shows last N servers
    --like LIKE              List action templates whose names contain LIKE
    --json-parameters JSON   Json parameters. Ex.: '{"folder":"/tmp",
    --json-target JSON       Target in json format. Ex.: '{"all":["node1", "node2"], "backend": "granule-backend"}'
    --name NAME              Action templates' name to list
    -p, --param PARAM=VALUE  Parameter passed to run orchestration. Examples: --parameter folder=/tmp
    -t, --target NAME=VALUE  Target to run the orchestrations NAME is the target name and value the granules or hosts
                             select. Example: --target all=node1,node2 --target backend=granule-backend
    --version N              Versions to list
    --no-wait                Does not wait the orchestration to finish
"""
import ast
import json
import re
import sys

from docopt import docopt

import dimensigon.dshell.network as ntwrk
from dimensigon.dshell.commands import orch_list, orch_run, dprint


def main(args):
    argv = docopt(__doc__, args)

    if argv['list']:
        orch_list(ident=argv['--id'], name=argv['--name'], version=argv['--version'], like=argv['--like'],
                  detail=argv['--detail'])
    elif argv['create']:
        data = None
        if argv['-']:
            content = sys.stdin.read()
            part = json.loads(content)
            data = part if isinstance(part, list) else [part]
        elif argv['FILE']:
            try:
                content = open(argv['FILE'], 'r').read()
            except Exception as e:
                sys.exit(str(e))
            try:
                data = json.loads(content)
            except Exception as e:
                exit(f"Json format error: {e}")
        else:
            exit("No file specified")
        if data:
            resp = ntwrk.post(f'api_1_0.orchestrations_full', json=data)
            dprint(resp)
        else:
            exit("No data found to create action template")
    elif argv['run']:
        orchestration_id = argv['ID']

        if argv['--json-parameters'] is None:
            container = {}
            for param in argv['--param']:
                match = re.search(r"^([\"']?)([\w\-_]+):(.*?)\1$", param, re.MULTILINE | re.DOTALL)
                if not match:
                    exit(
                        f"Not a valid parameter '{param}'. Must contain a KEY:VALUE. Ex. string-key:'string-value' or "
                        f"integer-key:12345 or list-key:[1,2]")
                else:
                    mark, key, value = match.groups()
                    try:
                        value = ast.literal_eval(value)
                    except Exception:
                        value = value.encode().decode('unicode-escape')
                if isinstance(value, str) and value.startswith('@'):
                    file = value.strip('@')
                    value = open(file, 'r').read()

                container.update({key: value})
            parameters = container
        else:
            parameters = json.loads(argv['--json-parameters'])

        if argv['--json-target'] is None:
            container = {}
            for value in argv['--target']:
                if '=' in value:
                    target, host = value.split('=')
                    if ',' in host:
                        host = host.split(',')
                    else:
                        host = [host]
                    if target in container:
                        container[target].extend(host)
                    else:
                        container.update({target: host})
                else:
                    if 'all' not in container:
                        container.update({'all': [value]})
                    else:
                        container['all'].append(value)
            target = container
        else:
            target = json.loads(argv['--json-target'])

        orch_run(orchestration_id, params=parameters, hosts=target, background=argv['--no-wait'] or False)



import typing as t

from dimensigon.domain.entities import User
from dimensigon.domain.entities.base import DistributedEntityMixin, SoftDeleteMixin
from dimensigon.utils.typos import UUID, Id
from dimensigon.web import db


class Vault(DistributedEntityMixin, SoftDeleteMixin, db.Model):
    __tablename__ = 'D_vault'

    user_id = db.Column(UUID, db.ForeignKey('D_user.id'), primary_key=True)
    scope = db.Column(db.String(100), primary_key=True, default='global')
    name = db.Column(db.String(100), primary_key=True)
    _old_name = db.Column(db.String(100))
    value = db.Column(db.PickleType)

    user = db.relationship("User")

    def to_json(self, human=False, **kwargs):
        dto = super().to_json(**kwargs)
        dto.update(name=self.name, value=self.value)
        if self.scope:
            dto.update(scope=self.scope)
        if human:
            dto.update(user={'id': self.user.id, 'name': self.user.name})
        else:
            dto.update(user_id=self.user_id or getattr(self.user, 'id', None))
        return dto

    @classmethod
    def from_json(cls, kwargs):
        super().from_json(kwargs)
        kwargs['user'] = User.query.get_or_raise(kwargs.pop('user_id', None))
        o = cls.query.get((getattr(kwargs.get('user'), 'id'), kwargs.get('scope', 'global'), kwargs.get('name')))
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)

    @classmethod
    def get_variables_from(cls, user: t.Union[Id, User], scope='global'):
        if isinstance(user, User):
            user_id = user.id
        else:
            user_id = user
        return {vault.name: vault.value for vault in cls.query.filter_by(user_id=user_id, scope=scope).all()}

    def __str__(self):
        return f"Vault({self.user}:{self.scope}[{self.name}={self.value}])"



import re

from setuptools import setup, find_packages

INIT_FILE = 'dimensigon/__init__.py'

with open("README.md", "r") as fh:
    long_description = fh.read()


def find_version():
    with open(INIT_FILE) as fp:
        for line in fp:
            # __version__ = '0.1.0'
            match = re.search(r"__version__\s*=\s*(['\"])([^\1]+)\1", line)
            if match:
                return match.group(2)
    assert False, 'cannot find version'


def find_author_email():
    with open(INIT_FILE) as fp:
        m_author, m_email = None, None
        for line in fp:
            if not m_author:
                m_author = re.search(r"__author__\s*=\s*(['\"])([^\1]*)\1", line)
            if not m_email:
                m_email = re.search(r"__email__\s*=\s*(['\"])([^\1]*)\1", line)
            if m_author and m_email:
                return m_author.group(2), m_email.group(2)
    assert False, 'cannot find author or email'


def find_licence():
    with open(INIT_FILE) as fp:
        for line in fp:
            match = re.search(r"__license__\s*=\s*(['\"])([^\1]*)\1", line)
            if match:
                return match.group(2)
    assert False, 'cannot find license'


def required_packages():
    with open('requirements.txt') as fp:
        return [line.strip() for line in fp if line.strip()]


author, email = find_author_email()
setup(
    name='dimensigon',
    version=find_version(),
    package_dir={"": "."},
    packages=find_packages(where=".", exclude=["contrib", "docs", "tests*", "tasks"]),
    url='https://github.com/dimensigon/dimensigon',
    license=find_licence(),
    author=author,
    author_email=email,
    description="Distributed Management and orchestration through RESTful, Mesh Networking and with a flair of IoT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    test_suite="tests",
    install_requires=required_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
    ],
    entry_points={'console_scripts': ["dshell=dimensigon.dshell.batch.dshell:main",
                                      "dimensigon=dimensigon.__main__:main"]},
    python_requires='>=3.6',
)



import copy
import datetime as dt
import functools
import inspect
import os
import re
import signal
import sqlite3
import sys
import tempfile
import time
import typing as t
from abc import ABC, abstractmethod

import flask
import jinja2
import requests
from dataclasses import dataclass
from flask import json
from packaging.version import parse

from dimensigon import defaults
from dimensigon.domain.entities import Step, Server, Software, OrchExecution, Orchestration, Scope, Route, StepExecution
from dimensigon.use_cases.lock import lock_scope
from dimensigon.utils import subprocess
from dimensigon.utils.helpers import get_now, is_iterable_not_string, format_exception, is_valid_uuid
from dimensigon.utils.typos import Kwargs
from dimensigon.utils.var_context import Context
from dimensigon.web import db, errors
from dimensigon.web import network as ntwrk
from dimensigon.web.helpers import normalize_hosts


@dataclass
class CompletedProcess:
    success: bool = None
    stdout: t.Union[str, bytes] = None
    stderr: t.Union[str, bytes] = None
    rc: int = None
    start_time: dt.datetime = None
    end_time: dt.datetime = None

    def set_start_time(self) -> 'CompletedProcess':
        self.start_time = get_now()
        return self

    def set_end_time(self) -> 'CompletedProcess':
        self.end_time = get_now()
        return self

    def to_json(self):
        return self.__dict__


class IOperationEncapsulation(ABC):

    def __init__(self, code: str, expected_stdout: str = None, expected_stderr: str = None, expected_rc: int = None,
                 system_kwargs: Kwargs = None):
        """
        Operation Initializer

        Parameters
        ----------
        code:
            Code to be executed
        expected_output:
            Expected output to compare from execution output. If None it won't be compared
        expected_rc:
            Expected return code to be compared with the execution return code. If None it won't be compared
        system_kwargs:
            system arguments to be passed to the commend execution, not the variables inside the code
        """
        self.code = code
        self.expected_stdout = expected_stdout
        self.expected_stderr = expected_stderr
        self.expected_rc = expected_rc
        self.system_kwargs = system_kwargs or {}

    def load_code(self):
        return

    @abstractmethod
    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        """
        StepExecution process
        Parameters
        ----------
        properties:
            properties passed to execute implementation
        env:
            environment variables

        Returns
        -------
        StepExecution:
            dataclass containing all the information from the result execution
        """

    def execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        """
        StepExecution wrapper
        Parameters
        ----------
        properties:
            properties passed to execute implementation
        env:
            environment variables

        Returns
        -------
        StepExecution:
            dataclass containing all the information from the result execution
        """
        start = get_now()
        try:
            return self._execute(params, timeout=timeout, context=context)
        except Exception as e:
            return CompletedProcess(success=False,
                                    stderr=f"Error creating execution: {format_exception(e)}",
                                    start_time=start,
                                    end_time=get_now())

    def rpl_params(self, **context):
        template = jinja2.Template(self.code)
        return template.render(**context)

    def evaluate_result(self, cp: CompletedProcess, context=None):
        if cp.success is None:
            res = []
            if self.expected_stdout is not None:
                if isinstance(cp.stdout, str):
                    match = re.search(self.expected_stdout, cp.stdout)
                    if match:
                        res.append(True)
                        if context:
                            for k, v in match.groupdict().items():
                                context.set(k, v)
                    else:
                        res.append(False)
            if self.expected_stderr is not None:
                if isinstance(cp.stderr, str):
                    match = re.search(self.expected_stderr, cp.stderr)
                    if match:
                        res.append(True)
                        if context:
                            for k, v in match.groupdict().items():
                                context.set(k, v)
                    else:
                        res.append(False)
            if self.expected_rc is not None:
                res.append(True) if self.expected_rc == cp.rc else res.append(False)
            cp.success = all(res)
        return cp


# class AnsibleOperation(IOperationEncapsulation):
# 
#     def execute(self, params: Kwargs, timeout=None, env=None) -> CompletedProcess:
#         code = self.code
#         if re.match(r'^[^<>:;,?"*|/\\]+$', self.code):
#             file = os.path.join('', 'ansible', self.code)
#             if os.path.exists(file):
#                 with open(file, 'r') as fh:
#                     code = fh.read()
# 
#         template = self.rpl_params({'globals': var_context.globals, **dict(var_context)})
# 
#         system_kwargs = self.system_kwargs.copy()
# 
#         tokens = ('ansible-playbook', '-i', '"localhost,"', '-c', 'local') + tuple(template)
# 
#         cp = CompletedProcess()
#         cp.set_start_time()
#         try:
#             r = subprocess.run(tokens, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
#                                **system_kwargs, timeout=timeout)
#             cp.stdout = r.stdout
#             cp.stderr = r.stderr
#             cp.rc = r.returncode
#         except (subprocess.TimeoutExpired, ValueError) as e:
#             cp.stderr = f"{e.__class__.__name__}{e.args}"
#             cp.success = False
#         finally:
#             cp.set_end_time()
# 
#         return self.evaluate_result(cp)

# TODO: make a test case
class RequestOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        kwargs = json.loads(self.rpl_params(params))
        kwargs.update(self.system_kwargs)
        cp = CompletedProcess()
        cp.set_start_time()

        # common parameters
        if kwargs.get('timeout') is not None and timeout is not None:
            kwargs['timeout'] = min(kwargs.get('timeout'), timeout)
        method = kwargs.pop('method').lower()

        resp, exception = None, None

        # process kwargs
        url = kwargs.pop('view_or_url', kwargs.pop('view', kwargs.pop('url', None)))

        # Run request
        try:
            resp = requests.request(method, url, **kwargs)
        except Exception as e:
            exception = e
        else:
            cp.rc = resp.status_code
            cp.stdout = resp.text
        if exception is None:
            self.evaluate_result(cp, context)
        else:
            cp.success = False
            cp.stderr = str(exception) if str(exception) else exception.__class__.__name__
        cp.set_end_time()
        return cp


class NativeOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None):
        pass


class NativeSoftwareSendOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        input_params = params['input']
        cp = CompletedProcess()
        cp.set_start_time()

        # common parameters
        kwargs = self.system_kwargs
        kwargs['timeout'] = timeout or kwargs.get('timeout')
        kwargs['identity'] = context.env.get('executor_id')

        resp, exception = None, None

        def search_cost(ssa, route_list):
            cost = [route['cost'] for route in route_list if str(ssa.server.id) == route['destination_id']]
            if cost:
                if cost[0] is None:
                    cost = 999999
                else:
                    cost = cost[0]
            else:
                cost = 999999
            return cost

        software = input_params.get('software', None)
        if is_valid_uuid(software):
            soft = Software.query.get(software)
            if not soft:
                cp.stderr = f"software id '{software}' not found"
                cp.success = False
                cp.set_end_time()
                return cp
        else:
            version = input_params.get('version', None)
            if version:
                parsed_ver = parse(str(version))
                soft_list = [s for s in Software.query.filter_by(name=software).all() if s.parsed_version == parsed_ver]
            else:
                soft_list = sorted(Software.query.filter_by(name=software).all(), key=lambda x: x.parsed_version)
            if soft_list:
                soft = soft_list[-1]
            else:
                cp.stderr = f"No software found for '{software}'" + (f" and version '{version}'" if version else "")
                cp.success = False
                cp.set_end_time()
                return cp

        if not soft.ssas:
            cp.stderr = f"{soft.id} has no server association"
            cp.success = False
            cp.set_end_time()
            return cp

        # Server validation
        server = input_params.get('server', None)
        if is_valid_uuid(server):
            dest_server = Server.query.get(server)
        else:
            dest_server = Server.query.filter_by(name=server).one_or_none()
        if not dest_server:
            cp.stderr = f"destination server {'id ' if is_valid_uuid(server) else ''}'{server}' not found"
            cp.success = False
            cp.set_end_time()
            return cp

        # decide best server source
        resp = ntwrk.get(dest_server, 'api_1_0.routes', timeout=10)
        if resp.code == 200:
            ssas = copy.copy(soft.ssas)
            ssas.sort(key=functools.partial(search_cost, route_list=resp.msg['route_list']))
        else:
            ssas = soft.ssas
        server = ssas[0].server

        # Process kwargs
        data = {'software_id': soft.id, 'dest_server_id': dest_server.id, "background": False,
                "include_transfer_data": True, "force": True}
        if input_params.get('dest_path', None):
            data.update(dest_path=input_params.get('dest_path', None))
        if input_params.get('chunk_size', None):
            data.update(chunk_size=input_params.get('chunk_size', None))
        if input_params.get('max_senders', None):
            data.update(max_senders=input_params.get('max_senders', None))
        # run request
        resp = ntwrk.post(server, 'api_1_0.send', json=data, **kwargs)
        cp.stdout = flask.json.dumps(resp.msg) if isinstance(resp.msg, dict) else resp.msg
        cp.stderr = str(resp.exception) if str(resp.exception) else resp.exception.__class__.__name__
        cp.rc = resp.code
        if resp.exception is None:
            self.evaluate_result(cp)
        cp.set_end_time()
        return cp


class NativeWaitOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None):
        input_params = params['input']
        start = time.time()
        found = []
        cp = CompletedProcess()
        cp.set_start_time()
        timeout = input_params.get('timeout')
        if timeout is None:
            timeout = self.system_kwargs.get('timeout')
        if timeout is None:
            timeout = defaults.MAX_TIME_WAITING_SERVERS

        try:
            now = get_now()
            server_names = input_params.get('server_names', [])
            if not is_iterable_not_string(server_names):
                server_names = [server_names]

            if not server_names:
                cp.success = False
                cp.stderr = f"No server to wait"
                cp.set_end_time()
                return cp

            pending_names = set(server_names)
            with lock_scope(Scope.CATALOG, retries=3, delay=4, applicant=context.env.get('orch_execution_id')):
                while len(pending_names) > 0:
                    try:
                        found_names = db.session.query(Server.name).filter(Server.name.in_(pending_names)).filter(
                            Server.created_on >= now).all()
                    except sqlite3.OperationalError as e:
                        if str(e) == 'database is locked':
                            found_names = []
                        else:
                            raise
                    found_names = set([t[0] for t in found_names]) if found_names else set()
                    pending_names = pending_names - found_names
                    if pending_names and (time.time() - start) < timeout:
                        time.sleep(self.system_kwargs.get('sleep_time', 15))
                    else:
                        break
        except errors.LockError as e:
            cp.success = False
            cp.stderr = str(e)

        else:
            if not pending_names:
                cp.success = True
                cp.stdout = f"Server{'s' if len(server_names) > 1 else ''} " \
                            f"{', '.join(sorted(server_names))} found"
            else:
                cp.success = False
                cp.stderr = f"Server{'s' if len(pending_names) > 1 else ''} {', '.join(sorted(pending_names))} " \
                            f"not created after {timeout} seconds"
        cp.set_end_time()
        return cp

# TODO: make a test case
class NativeDmRunningOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None):
        input_params = params['input']
        start = time.time()
        running = []
        cp = CompletedProcess()
        cp.set_start_time()
        try:
            min_timeout = min(timeout, input_params.get('timeout', None))
        except TypeError:
            if timeout is not None:
                min_timeout = timeout
            elif input_params.get('timeout', None) is not None:
                min_timeout = input_params.get('timeout')
            else:
                min_timeout = defaults.MAX_TIME_WAITING_SERVERS

        server_names = input_params.get('server_names', [])
        if not is_iterable_not_string(server_names):
            server_names = [server_names]

        if not server_names:
            cp.success = False
            cp.stdout = f"No server to wait for DM running"
            cp.set_end_time()
            return cp
        pending_names = set(server_names)
        found_names = []
        while len(pending_names) > 0:
            try:
                found_names = db.session.query(Server.name).join(Route, Route.destination_id == Server.id).filter(
                    Server.name.in_(pending_names)).filter(Route.cost.isnot(None)).order_by(Server.name).all()
            except sqlite3.OperationalError as e:
                if str(e) == 'database is locked':
                    pass
                else:
                    raise
            found_names = set([t[0] for t in found_names])
            pending_names = pending_names - found_names
            if pending_names and (time.time() - start) < min_timeout:
                time.sleep(self.system_kwargs.get('sleep_time', 15))
            else:
                break

        if not pending_names:
            cp.success = True
            cp.stdout = f"Server{'s' if len(server_names) > 1 else ''} " \
                        f"{', '.join(sorted(server_names))} with dimensigon running"
        else:
            cp.success = False
            cp.stderr = f"Server{'s' if len(pending_names) > 1 else ''} {', '.join(sorted(pending_names))} " \
                        f"with dimensigon not running after {min_timeout} seconds"
        cp.set_end_time()
        return cp

# TODO: make a test case
class NativeDeleteOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None):
        input_params = params['input']
        cp = CompletedProcess()
        cp.set_start_time()
        try:
            with lock_scope(Scope.CATALOG, retries=3, delay=4, applicant=context.env.get('orch_execution_id')):
                server_names = input_params.get('server_names', [])
                if not is_iterable_not_string(server_names):
                    server_names = [server_names]
                to_be_deleted = Server.query.filter(Server.name.in_(server_names)).all()

                for s in to_be_deleted:
                    # remove associated routes
                    db.session.delete(s.route)
                    s.delete()
                    db.session.commit()

        except errors.LockError as e:
            cp.success = False
            cp.stderr = str(e)
        except Exception as e:
            cp.success = False
            cp.stderr = f"Unable to delete server{'s' if len(input_params.get('to_be_deleted', [])) > 1 else ''} " \
                        f"{', '.join([s.name for s in to_be_deleted])}. Exception: {e}"
        else:
            cp.success = True
            cp.stdout = f"Server{'s' if len(to_be_deleted) > 1 else ''} " \
                        f"{', '.join([s._old_name for s in to_be_deleted])} deleted"

        cp.set_end_time()
        return cp


class ShellOperation(IOperationEncapsulation):

    def _run(self, code, user=None, shebang=None, timeout=None):
        stdout = None
        stderr = None
        rc = None
        tmp = tempfile.NamedTemporaryFile('w', delete=False, suffix='.' + os.path.basename(shebang))
        tmp.write(f"#!{shebang}\n")
        tmp.write(code)
        tmp.close()
        os.chmod(tmp.name, 0o755)
        out_fh = open(tmp.name + '.out', 'w')

        if user:
            cmd = f"sudo -niu {user} {tmp.name}"
        else:
            cmd = tmp.name

        with subprocess.Popen(cmd, stdout=out_fh, stderr=subprocess.STDOUT, encoding='utf-8',
                              shell=True, env=os.environ) as p:
            try:
                p.communicate(timeout=timeout)
            except subprocess.TimeoutExpired:
                subprocess.run(f"sudo -u {user} kill {p.pid}", shell=True) if user else p.terminate()
                try:
                    p.communicate(timeout=5)
                except subprocess.TimeoutExpired:
                    subprocess.run(f"sudo -u {user} kill -9 {p.pid}", shell=True) if user else os.kill(p.pid,
                                                                                                       signal.SIGKILL)
                    p.communicate()
                    raise TimeoutError(f"Timeout of {timeout} seconds while executing shell")
                except Exception as e:
                    subprocess.run(f"sudo -u {user} kill -9 {p.pid}", shell=True) if user else os.kill(p.pid,
                                                                                                       signal.SIGKILL)
                    raise RuntimeError(f"Error waiting process {p.pid} to terminate\n{format_exception(e)}")
                else:
                    raise TimeoutError(f"Timeout of {timeout} seconds while executing shell")
            except Exception as e:
                subprocess.run(f"sudo -u {user} kill -9 {p.pid}", shell=True) if user else os.kill(p.pid,
                                                                                                   signal.SIGKILL)
                p.communicate()
                stderr += "\n" + format_exception(e)

            rc = p.poll()
            out_fh.close()
            with open(tmp.name + '.out', 'r') as fh:
                stdout = fh.read()

        try:
            os.remove(tmp.name)
        except:
            pass

        try:
            os.remove(tmp.name + '.out')
        except:
            pass

        return stdout, stderr, rc

    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        tokens = self.rpl_params(**params, env=context.env)

        system_kwargs = self.system_kwargs.copy()

        shebang = system_kwargs.pop('shebang', '/bin/bash')
        user = system_kwargs.pop('run_as', None)
        timeout = system_kwargs.pop('timeout', timeout)
        if timeout and not isinstance(timeout, float):
            try:
                timeout = float(timeout)
            except:
                timeout = None
        cp = CompletedProcess()
        cp.set_start_time()
        try:
            cp.stdout, cp.stderr, cp.rc = self._run(tokens, user, shebang, timeout)
        except Exception as e:
            cp.stderr = str(e)
            cp.success = False
        cp.set_end_time()

        self.evaluate_result(cp, context)

        return cp

# TODO: make a test case
class OrchestrationOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        from dimensigon.use_cases.deployment import deploy_orchestration
        input_params = params['input']
        cp = CompletedProcess().set_start_time()
        # Validation
        orchestration = Orchestration.get(input_params.pop('orchestration', None), input_params.pop('version', None))
        if not isinstance(orchestration, Orchestration):
            cp.stderr = orchestration
            cp.success = False
            return cp.set_end_time()

        hosts = input_params.pop('hosts', None)
        if hosts is None:
            cp.stderr = "No hosts specified"
            cp.success = False
            return cp.set_end_time()
        hosts = copy.deepcopy(hosts)
        if isinstance(hosts, list):
            hosts = {'all': hosts}
        elif isinstance(hosts, str):
            hosts = {'all': [hosts]}
        not_found = normalize_hosts(hosts)
        if not_found:
            cp.stderr = str(errors.ServerNormalizationError(not_found))
            cp.success = False
            return cp.set_end_time()

        o_exe = OrchExecution(orchestration_id=orchestration.id,
                              target=hosts,
                              params=input_params,
                              executor_id=context.env.get('executor_id'),
                              server_id=context.env.get('server_id'),
                              parent_step_execution_id=context.env.get('step_execution_id'))
        db.session.add(o_exe)
        se = StepExecution.query.get(context.env.get('step_execution_id'))
        if se:
            se.child_orch_execution_id = o_exe.id
        db.session.commit()

        cp.stdout = f"orch_execution_id={o_exe.id}"

        env = dict(context.env)
        env.update(root_orch_execution_id=context.env['root_orch_execution_id'],
                   orch_execution_id=o_exe.id,
                   executor_id=context.env['executor_id'])
        ctx = Context(input_params, env, vault=context.vault)
        try:
            deploy_orchestration(orchestration=orchestration, hosts=hosts, var_context=ctx, execution=o_exe,
                                 timeout=timeout)
        except Exception as e:
            cp.stderr = str(e) if str(e) else e.__class__.__name__
            cp.success = False
        else:
            context.merge_ctx(ctx)
            db.session.refresh(o_exe)
            cp.success = o_exe.success
        return cp.set_end_time()


class TestOperation(IOperationEncapsulation):

    def _execute(self, params: Kwargs, timeout=None, context: Context = None) -> CompletedProcess:
        tokens = self.rpl_params(**params, env=context.env)
        return CompletedProcess(success=True, stdout=tokens, rc=0, start_time=get_now(), end_time=get_now())

from dimensigon.domain.entities import ActionType

_operation_classes = {}
for name, cls in inspect.getmembers(sys.modules['dimensigon.use_cases.operations'],
                                    lambda x: (inspect.isclass(x) and issubclass(x, IOperationEncapsulation))):
    _operation_classes.update({name: cls})

_factories: t.Dict['ActionType', t.Type[IOperationEncapsulation]] = {}

for at in ActionType:
    try:
        _factories.update({at: _operation_classes[at.name.capitalize() + 'Operation']})
    except KeyError:
        NotImplementedError(f"{at.name.capitalize() + 'Operation'} not implemented")


def create_operation(step: 'Step') -> IOperationEncapsulation:
    kls = _factories[step.action_type]

    if kls == NativeOperation:
        if step.action_template.id == '00000000-0000-0000-000a-000000000001':
            kls = NativeSoftwareSendOperation
        elif step.action_template.id == '00000000-0000-0000-000a-000000000002':
            kls = NativeWaitOperation
        elif step.action_template.id == '00000000-0000-0000-000a-000000000004':
            kls = NativeDmRunningOperation
        elif step.action_template.id == '00000000-0000-0000-000a-000000000005':
            kls = NativeDeleteOperation
    return kls(code=step.code, expected_stdout=step.expected_stdout, expected_stderr=step.expected_stderr,
               expected_rc=step.expected_rc,
               system_kwargs=step.system_kwargs)



import time
import typing as t

from flask import Flask, g
from flask_jwt_extended import JWTManager
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData
from sqlalchemy.pool import StaticPool

from dimensigon.utils.event_handler import EventHandler
from dimensigon.web import errors, threading
from dimensigon.web.config import config_by_name
from .extensions.flask_executor.executor import Executor
from .helpers import BaseQueryJSON, run_in_background, get_root_auth
from ..utils.helpers import bind2gate

if t.TYPE_CHECKING:
    from ..core import Dimensigon

meta = MetaData(naming_convention={
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_N_key)s",
    "ck": "ck_%(table_name)s_%(column_0_N_key)s",
    "fk": "fk_%(table_name)s_%(column_0_N_key)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
})

db = SQLAlchemy(query_class=BaseQueryJSON, metadata=meta,
                engine_options=dict(connect_args={'check_same_thread': False}, isolation_level='READ UNCOMMITTED'))
# db = SQLAlchemy(query_class=BaseQueryJSON, metadata=meta)
# db = SQLAlchemy(query_class=BaseQueryJSON)
jwt = JWTManager()
executor = Executor()


class DimensigonFlask(Flask):
    dm: t.ClassVar['Dimensigon'] = None

    def bootstrap(self):
        """ bootstraps the application. Gunicorn is still not listening on sockets
        """
        with self.app_context():
            from dimensigon.domain.entities import Server, Parameter
            import dimensigon.web.network as ntwrk
            from dimensigon.domain.entities import Locker

            # reset scopes
            Locker.set_initial(unlock=True)

            # check gates
            me = Server.get_current()
            if me is None:
                raise RuntimeError("No server set as 'current'")

            input_gates = bind2gate(self.dm.config.http_conf.get('bind'))
            current_gates = [(gate.dns or str(gate.ip), gate.port) for gate in me.gates]
            new_gates = set(input_gates).difference(set(current_gates))
            self.server_id_with_new_gates = None
            if new_gates:
                if Parameter.get('join_server'):
                    join_server = Server.query.get(Parameter.get('join_server'))
                else:
                    join_server = None
                servers = Server.get_neighbours()
                if join_server in servers:
                    servers.pop(servers.index(join_server))
                    servers.append(join_server)
                else:
                    self.logger.warning(f'Join server {join_server} is not a neighbour')
                start = time.time()
                resp = None
                server = True
                while len(servers) > 0 and server and (time.time() - start) < 900:
                    server_retries = 0
                    server = servers[-1]
                    self.logger.debug(f"Sending new gates {new_gates} to {server}...")
                    resp = ntwrk.patch(server, 'api_1_0.serverresource',
                                       view_data=dict(server_id=str(Server.get_current().id)),
                                       json={'gates': [{'dns_or_ip': ip, 'port': port} for ip, port in new_gates]},
                                       timeout=60,
                                       auth=get_root_auth())
                    if not resp.ok:
                        self.logger.debug(f"Unable to send new gates to {server}. Reason: {resp}")
                        self.logger.info(f"Unable to create new gates. Trying to send again in 5 seconds...")
                        time.sleep(5)
                        if resp.code == 409:
                            # try with the same server
                            server_retries += 1
                        elif resp.code == 500:

                            # try with another server
                            i = servers.index(server) - 1
                            if i >= 0:
                                server = servers[i]
                                server_retries = 0
                            else:
                                server = None
                        if server_retries == 3:
                            # changing server
                            i = servers.index(server) - 1
                            if i >= 0:
                                server = servers[i]
                                server_retries = 0
                            else:
                                server = None
                    else:
                        self.logger.debug("New gates created succesfully")
                        Parameter.set('new_gates_server', server.id)
                        break

                if not servers:
                    if Server.query.count() == 1:
                        self.logger.info(f"Creating new gates {new_gates} without performing a lock on catalog")
                        for gate in new_gates:
                            g = me.add_new_gate(gate[0], gate[1])
                            db.session.add(g)

                else:
                    if resp and not resp.ok:
                        self.logger.warning(f"Remote servers may not connect with {me}. ")
                db.session.commit()

    def run(self, host=None, port=None, debug=None, **options):
        super(DimensigonFlask, self).run(host=host, port=port, debug=debug, use_reloader=False, **options)


def _initialize_blueprint(app):
    from dimensigon.web.routes import root_bp
    from dimensigon.web.api_1_0 import api_bp as api_1_0_bp

    app.register_blueprint(root_bp)
    handle_exception = app.handle_exception
    handle_user_exception = app.handle_user_exception
    app.register_blueprint(api_1_0_bp)
    app.handle_exception = handle_exception
    app.handle_user_exception = handle_user_exception


def _initialize_errorhandlers(app):
    '''
    Initialize error handlers
    '''
    from dimensigon.web.errors import bp_errors
    app.register_blueprint(bp_errors)


def create_app(config_name):
    app = DimensigonFlask('dm')
    if isinstance(config_name, t.Mapping):
        app.config.from_mapping(config_name)
    elif config_name in config_by_name:
        app.config.from_object(config_by_name[config_name])
        config_by_name[config_name].init_app(app)
    else:
        app.config.from_object(config_name)
        if hasattr(config_name, 'init_app'):
            config_name.init_app(app)

    # EXTENSIONS
    db.init_app(app)
    jwt.init_app(app)
    executor.init_app(app)
    app.events = EventHandler()

    # with app.app_context():
    #     def do_connect(dbapi_connection, connection_record):
    #         # disable pysqlite's emitting of the BEGIN statement entirely.
    #         # also stops it from emitting COMMIT before any DDL.
    #         dbapi_connection.isolation_level = None
    #
    #     def do_begin(conn):
    #         # emit our own BEGIN
    #         conn.execute("BEGIN")
    #     event.listen(db.get_engine(), "connect", do_connect)
    #     event.listen(db.get_engine(), "begin", do_begin)

    app.before_request(load_global_data_into_context)
    # if not app.config['TESTING']:
    # app.before_first_request(app.dm.cluster_manager.notify_cluster)
    # app.before_first_request(app.cluster_manager.start)
    # app.before_first_request(app.file_sync.start)
    _initialize_blueprint(app)
    _initialize_errorhandlers(app)

    return app


# @jwt.user_loader_callback_loader
# def user_loader_callback(identity):
#     from ..domain.entities import User
#     return User.query.get(identity)


def load_global_data_into_context():
    from dimensigon.domain.entities import Server, Dimension
    from dimensigon.web.decorators import set_source
    global _dimension, _server
    set_source()
    g.server = Server.get_current()
    g.dimension = Dimension.get_current()



import datetime as dt
import json
import multiprocessing as mp
import time
import typing as t

from pkg_resources import parse_version
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.attributes import flag_modified

from dimensigon import __version__
from dimensigon import defaults
from dimensigon.domain.entities import bypass_datamark_update, Scope, Server, Catalog
from dimensigon.use_cases import mptools as mpt
from dimensigon.use_cases.lock import lock_scope
from dimensigon.use_cases.mptools import TerminateInterrupt
from dimensigon.utils import asyncio
from dimensigon.utils.helpers import get_distributed_entities, get_now
from dimensigon.web import errors, db, get_root_auth
from dimensigon.web import network as ntwrk

if t.TYPE_CHECKING:
    from dimensigon.core import Dimensigon


class BaseCatalogException(errors.BaseError):
    """Base Class Exception"""


class NewVersionFound(BaseCatalogException):
    """Exception raised if new version found"""


class NoServerFound(BaseCatalogException):
    """No server with high catalog found"""


class NoNeighbourHealtchcheckFound(BaseCatalogException):
    """No healthcheck found server with high catalog found"""


class CatalogFetchError(BaseCatalogException):
    """Unable to fetch catalog from server"""


class HealthCheckMismatch(errors.BaseError):
    status_code = 500

    def __init__(self, expected: t.Dict[str, str], actual: t.Dict[str, str]):
        self.expected = expected
        self.actual = actual

    def _format_error_msg(self) -> str:
        return "Healtcheck response does not match with the server requested"


class CatalogMismatch(errors.BaseError):
    status_code = 500

    def __init__(self, local_entities, remote_entities):
        self.local_entities = local_entities
        self.remote_entities = remote_entities

    def _format_error_msg(self) -> str:
        return "List entities do not match"


def update_db_catalog(catalog, check_mismatch=True, logger=None):
    de = get_distributed_entities()

    if check_mismatch:
        inside = set([name for name, cls in de])
        logger.debug(f'Actual entities: {inside}') if logger else None

        outside = set(catalog.keys())
        logger.debug(f'Remote entities: {outside}') if logger else None

        if len(inside ^ outside) > 0:
            raise CatalogMismatch(inside, outside)

    db.session.flush()
    with bypass_datamark_update():
        for name, cls in de:
            if name in catalog:
                if len(catalog[name]) > 0:
                    logger.log(1, f"Adding/Modifying new '{name}' entities: \n"
                                  f"{json.dumps(catalog[name], indent=2, sort_keys=True)}") if logger else None
                for dto in catalog[name]:
                    o = cls.from_json(dict(dto))
                    # force modification to update catalog last_modified_at
                    flag_modified(o, 'last_modified_at')
                    db.session.add(o)

        db.session.commit()


class CatalogManager(mpt.TimerWorker):
    INTERVAL_SECS = 5 * 60

    ###########################
    # START Class Inheritance #
    def init_args(self, dimensigon: 'Dimensigon', interval_secs=defaults.CATALOG_REFRESH_PERIOD):
        self.dm = dimensigon
        self.Session = sessionmaker(bind=self.dm.engine)
        self.INTERVAL_SECS = interval_secs
        self._catalog_ver = None
        self._updating = mp.Event()
        self._update_lock = mp.Lock()
        self._server = None

    def main_func(self):
        with self._update_lock:
            with self.dm.flask_app.app_context():
                self._updating.set()
                try:
                    self.upgrade_process()
                except NewVersionFound as e:
                    self.logger.info(f'Found mayor version on server {e.args[0]}. Upgrade version first')
                except NoServerFound as e:
                    pass
                except CatalogFetchError as e:
                    self.logger.error(f'Error fetching catalog from {e.args[0].server}: {e.args[0]}')
                except (KeyboardInterrupt, TerminateInterrupt):
                    pass
                except Exception as e:
                    self.logger.exception("Exception while trying to execute catalog update")
                finally:
                    self._updating.clear()

    # END Class Inheritance #
    #########################

    ############################
    # INIT Interface functions #
    def force_catalog_update(self):
        if self._updating.is_set():
            return None
        else:
            self.next_time = None
            self.main_func()
            self.next_time = time.time() + self.INTERVAL_SECS
            return self.catalog_ver

    # END Interface functions  #
    ############################

    ##############################
    # INNER methods & attributes #

    # @property
    # def session(self):
    #     if not hasattr(self, '_session') or self._session is None:
    #         self._session = self.Session()
    #     return self._session

    def upgrade_process(self):
        self.logger.debug("Starting check catalog from neighbours")
        # cluster information
        cluster_hearthbeat_id = get_now().strftime(defaults.DATETIME_FORMAT)
        # check version update before catalog update to match database revision
        data = asyncio.run(self._async_get_neighbour_healthcheck(cluster_hearthbeat_id))
        if data:
            self.check_new_version(data)
            self.catalog_update(data)
        else:
            raise NoServerFound()

    @property
    def server(self) -> Server:
        if self._server is None:
            self._server = Server.query.get(self.dm.server_id)
        return self._server

    @property
    def catalog_ver(self) -> dt.datetime:
        return db.session.query(db.func.max(Catalog.last_modified_at)).scalar()

    async def _async_get_neighbour_healthcheck(self, cluster_heartbeat_id: str = None) -> t.Dict[
        Server, dict]:

        server_responses = {}
        servers = Server.get_neighbours()
        self.logger.debug(f"Neighbour servers to check: {', '.join([s.name for s in servers])}")

        auth = get_root_auth()
        if cluster_heartbeat_id is None:
            cluster_heartbeat_id = get_now().strftime(defaults.DATETIME_FORMAT)

        cos = [ntwrk.async_post(server, 'root.healthcheck',
                                json={'me': self.dm.server_id,
                                      'heartbeat': cluster_heartbeat_id},
                                auth=auth) for server in servers]
        responses = await asyncio.gather(*cos)
        for server, resp in zip(servers, responses):
            if resp.ok:
                id_response = resp.msg.get('server', {}).get('id', '')
                if id_response and str(server.id) != id_response:
                    e = HealthCheckMismatch(
                        expected={'id': str(server.id), 'name': server.name},
                        actual=resp.msg.get('server', {}))
                    self.logger.warning(str(e))
                else:
                    server_responses.update({server: resp.msg})
            else:
                self.logger.warning(f"Unable to get Healthcheck from server {server.name}: {resp}")
        return server_responses

    def check_new_version(self, data: t.Dict[Server, dict]):
        mayor_version, mayor_server = None, None
        for server, hc in data.items():
            remote_version = parse_version(hc['version'])
            if remote_version > parse_version(__version__):
                if mayor_version is None or mayor_version < remote_version:
                    mayor_version, mayor_server = remote_version, server
        if mayor_version:
            raise NewVersionFound(str(mayor_server))

    def catalog_update(self, data: t.Dict[Server, dict]):
        reference_server = None
        if self.catalog_ver:
            for server, hc_msg in data.items():

                new_catalog_ver = dt.datetime.strptime(hc_msg['catalog_version'],
                                                       defaults.DATEMARK_FORMAT)
                if new_catalog_ver > self.catalog_ver:
                    reference_server = server
            if reference_server:
                self.logger.info(f"New catalog found from server {reference_server.name}: {self.catalog_ver}")
                self._update_catalog_from_server(reference_server)
            else:
                raise NoServerFound()

    def _update_catalog_from_server(self, server):
        with lock_scope(Scope.UPGRADE, [self.server]):
            resp = ntwrk.get(server, 'api_1_0.catalog',
                             view_data=dict(data_mark=self.catalog_ver.strftime(defaults.DATEMARK_FORMAT)),
                             auth=get_root_auth())

            if resp.code and 199 < resp.code < 300:
                delta_catalog = resp.msg
                self.db_update_catalog(delta_catalog)
            else:
                raise CatalogFetchError(resp)

    def db_update_catalog(self, catalog, check_mismatch=True):
        update_db_catalog(catalog, check_mismatch=check_mismatch, logger=self.logger)



import asyncio
import datetime as dt
import json
import logging
import random
import threading
import time
import typing as t
import uuid

from sqlalchemy import orm
from sqlalchemy.orm import sessionmaker

from dimensigon import defaults
from dimensigon.domain.entities import Server, Route, Gate, Parameter
from dimensigon.domain.entities.route import RouteContainer
from dimensigon.network.low_level import check_host, async_check_host
from dimensigon.use_cases.mptools import Worker, MPQueue
from dimensigon.use_cases.mptools_events import BaseEvent
from dimensigon.utils.helpers import convert, is_iterable_not_string, format_exception, get_now
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, errors, get_root_auth

if t.TYPE_CHECKING:
    from dimensigon.core import Dimensigon

MAX_COST = 99999


class RouteEvent(BaseEvent):
    """Route related Event"""


class InitialRouteSet(RouteEvent):
    """Event published when route inital process has completed"""


def check_gates(server, timeout=5, retries=1, delay=0.5) -> t.Union[RouteContainer, Route]:
    """checks if a server is neighbour

    Args:
        server: server to discover as a new neighbour
        timeout: timeout socket
        retries: number of times to try to connect to the socket

    Returns:
        returns Route if no change. This means that gate in inventory is still up.
                RouteContainer if there is an available gate to connect
                None if no gate to connect
    """

    # check current gate
    if server.route and server.route.gate:
        gate = server.route.gate
        if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=delay, timeout=timeout):
            return server.route

    for gate in server.gates:
        if server.route and gate != server.route.gate:
            if (gate.ip and not gate.ip.is_loopback) or (gate.dns and gate.dns != 'localhost'):
                if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=1, timeout=timeout):
                    return RouteContainer(None, gate, 0)


async def async_check_gates(server, timeout=5, retries=1, delay=1) -> t.Union[RouteContainer, Route]:
    """checks if a server is neighbour

    Args:
        server: server to discover as a new neighbour
        timeout: timeout socket
        retries: number of times to try to connect to the socket
        delay: time in seconds between retries

    Returns:
        returns Route if no change. This means that gate in inventory is still up.
                RouteContainer if there is an available gate to connect
                None if no gate to connect
    """

    # check current gate
    if server.route and server.route.gate:
        gate = server.route.gate
        if await async_check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries + 1, delay=delay,
                                  timeout=timeout):
            return server.route

    for gate in server.gates:
        if server.route and gate != server.route.gate:
            if (gate.ip and not gate.ip.is_loopback) or (gate.dns and gate.dns != 'localhost'):
                if await async_check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries,
                                          delay=1,
                                          timeout=timeout):
                    return RouteContainer(None, gate, 0)


async def _async_set_current_neighbours(neighbours: t.List[Server] = None,
                                        changed_routes: t.Dict[Server, RouteContainer] = None) -> t.List[Server]:
    """Function checks and sets neighbours

    Args:
        neighbours: list of neighbours
        changed_routes: reference to a dict which will be populated with new routes

    Returns:
        list of servers which are not neighbours anymore
    """
    not_neighbours_anymore = []

    if neighbours is None:
        neighbours = Server.get_neighbours()

    if neighbours:
        resp = await asyncio.gather(*[async_check_gates(server) for server in neighbours])
        for route, server in zip(resp, neighbours):
            if isinstance(route, RouteContainer):
                server.set_route(route)
                if changed_routes is not None:
                    changed_routes[server] = route
            elif route is None:
                not_neighbours_anymore.append(server)
                rc = RouteContainer(None, None, None)
                server.set_route(rc)
                if changed_routes is not None:
                    changed_routes[server] = rc
    return not_neighbours_anymore


async def _async_discover_new_neighbours(servers: t.List[Server] = None,
                                         changed_routes: t.Dict[Server, RouteContainer] = None) -> t.List[Server]:
    new_neighbours = []

    if servers is None:
        servers = Server.get_not_neighbours()

    if servers:
        resp = await asyncio.gather(*[async_check_gates(server) for server in servers], return_exceptions=True)
        for new_route, server in zip(resp, servers):
            if new_route:
                server.set_route(new_route)
                new_neighbours.append(server)
                changed_routes[server] = new_route

    return new_neighbours


class RouteManager(Worker):
    ###########################
    # START Class Inheritance #
    def init_args(self, dimensigon: 'Dimensigon', maxsize=None, refresh_interval=defaults.ROUTE_REFRESH_PERIOD,
                  send_interval=defaults.ROUTE_SEND_PERIOD):
        self.dm = dimensigon
        self.Session = sessionmaker(bind=self.dm.engine)
        self.queue = MPQueue(maxsize=maxsize or 10000)
        self._changed_routes: t.Dict[Id, t.Dict] = {}
        self.refresh_interval = refresh_interval
        self.send_interval = send_interval
        self._loop = asyncio.new_event_loop()

    def startup(self):
        def refresh():
            self.refresh_table(discover_new_neighbours=True, check_current_neighbours=True, max_num_discovery=5)
            self._timer = threading.Timer(interval=self.refresh_interval, function=refresh)
            self._timer.start()

        self.session = self._create_session()
        self._timer = threading.Timer(interval=self.refresh_interval, function=refresh)
        self._timer.start()
        self._next_send = time.time() + self.send_interval

    def shutdown(self):
        self.queue.close()
        self.queue.join_thread()
        self.session.close()
        if self._timer:
            self._timer.cancel()
            self._timer.join()

    def _main_loop(self):
        with self.dm.flask_app.app_context():
            last_shutdown = Parameter.get('last_graceful_shutdown')

            try:
                last_shutdown = dt.datetime.strptime(last_shutdown, defaults.DATETIME_FORMAT)
            except:
                last_shutdown = get_now()
            else:
                last_shutdown = last_shutdown

            if self.dm.config.force_scan or last_shutdown < (get_now() - dt.timedelta(seconds=self.refresh_interval)):
                scan = True
            else:
                scan = False
            changed_routes = self._loop.run_until_complete(
                self._async_refresh_route_table(discover_new_neighbours=scan, check_current_neighbours=scan,
                                                max_num_discovery=None))
            self.session.commit()
            self.publish_q.safe_put(InitialRouteSet())

            super()._main_loop()

            Parameter.set('last_graceful_shutdown', get_now().strftime(defaults.DATETIME_FORMAT))

    def main_func(self, *args, **kwargs):
        item = self.queue.safe_get()
        try:
            if item:

                if isinstance(item, tuple) and item[0] == 'REFRESH':
                    changed_routes = self._loop.run_until_complete(self._async_refresh_route_table(*item[1:]))
                elif isinstance(item, tuple) and item[0] == 'NEW':
                    changed_routes = self._new_node_in_cluster(*item[1:])
                elif isinstance(item, tuple) and item[0] == 'REMOVE':
                    changed_routes = self._remove_node_from_cluster(*item[1:])
                else:
                    changed_routes = self._update_route_table_from_data(item)
                if changed_routes:
                    self.session.commit()
                self._changed_routes.update(changed_routes)
            if time.time() > self._next_send:
                if self._changed_routes:
                    self._loop.run_until_complete(self._send_routes())
                self._next_send = time.time() + self.send_interval
        except Exception as e:
            self.logger.exception(f"Error processing item {item}")
            self.session.rollback()

    # END Class Inheritance #
    #########################

    ############################
    # INIT Interface functions #
    def refresh_table(self, discover_new_neighbours, check_current_neighbours, max_num_discovery):
        self.queue.safe_put(('REFRESH', discover_new_neighbours, check_current_neighbours, max_num_discovery))

    def new_node_in_cluster(self, server_id, routes):
        self.queue.safe_put(('NEW', server_id, routes))

    def remove_node_from_cluster(self, server_id):
        self.queue.safe_put(('REMOVE', server_id))

    def new_routes(self, new_routes):
        self.queue.safe_put(new_routes)

    # END Interface functions  #
    ############################

    ##############################
    # INNER methods & attributes #
    def _create_session(self):
        self.Session = sessionmaker(bind=self.dm.engine, autoflush=False)
        return self.Session()

    @property
    def server(self) -> Server:
        if getattr(self, '_server', None) is None:
            self._server = self.session.query(Server).filter_by(_me=1, deleted=0).one_or_none()
        return self._server

    @property
    def server_query(self):
        return self.session.query(Server).filter_by(deleted=0) if self.session else None

    @property
    def gate_query(self):
        return self.session.query(Gate).filter_by(deleted=0) if self.session else None

    def _route_table_merge(self, data: t.Dict[Server, ntwrk.Response]):
        changed_routes: t.Dict[Server, RouteContainer] = {}
        temp_table_routes: t.Dict[uuid.UUID, t.List[RouteContainer]] = {}
        for s, resp in data.items():
            if resp.code == 200:
                server_id = resp.msg.get('server_id', None) or resp.msg.get('server').get('id')
                likely_proxy_server_entity = self.session.query(Server).get(server_id)
                for route_json in resp.msg['route_list']:
                    route_json = convert(route_json)
                    if route_json.destination_id != self.server.id \
                            and route_json.proxy_server_id != self.server.id \
                            and route_json.gate_id not in [g.id for g in self.server.gates]:
                        if route_json.destination_id not in temp_table_routes:
                            temp_table_routes.update({route_json.destination_id: []})
                        if route_json.cost is not None:
                            route_json.cost += 1
                            route_json.proxy_server_id = likely_proxy_server_entity.id
                            route_json.gate_id = None
                            temp_table_routes[route_json.destination_id].append(
                                RouteContainer(likely_proxy_server_entity.id, None, route_json.cost))
                        elif route_json.cost is None:
                            # remove a routing if gateway cannot reach the destination
                            temp_table_routes[route_json.destination_id].append(
                                RouteContainer(route_json.proxy_server_id, None, None))
            else:
                self.logger.error(f"Error while connecting with {s}. Error: {resp}")

        # Select new routes based on neighbour routes
        neighbour_ids = [s.id for s in Server.get_neighbours(session=self.session)]
        for destination_id in filter(lambda s: s not in neighbour_ids, temp_table_routes.keys()):
            route = self.session.query(Route).filter_by(destination_id=destination_id).one_or_none()
            if not route:
                server = self.session.query(Server).get(destination_id)
                if not server:
                    continue
                else:
                    route = Route(destination=server)
            temp_table_routes[destination_id].sort(key=lambda x: x.cost or MAX_COST)
            if len(temp_table_routes[destination_id]) > 0:
                min_route = temp_table_routes[destination_id][0]
                proxy_server: Server = self.session.query(Server).get(min_route.proxy_server)
                cost = min_route.cost
                if route.proxy_server != proxy_server or route.cost != cost:
                    rc = RouteContainer(proxy_server, None, cost)
                    route.set_route(rc)
                    changed_routes[route.destination] = rc
                    self.session.add(route)

        data = {}
        for server, temp_route in changed_routes.items():
            data.update({str(server): {'proxy_server': str(temp_route.proxy_server), 'gate': str(temp_route.gate),
                                       'cost': str(temp_route.cost)}})
        return changed_routes

    async def _async_refresh_route_table(self, discover_new_neighbours=False, check_current_neighbours=False,
                                         max_num_discovery=None) -> t.Dict[Server, RouteContainer]:
        """Gets route tables of all neighbours and updates its own table based on jump weights.
        Needs a Flask App Context to run.

        Parameters
        ----------
        discover_new_neighbours:
            tries to discover new neighbours
        check_current_neighbours:
            checks if current neighbours are still neighbours
        max_num_discovery:
            maximum number of possible nodes to check as neighbour

        Returns
        -------
        None
        """

        self.logger.debug('Refresh Route Table')
        neighbours = Server.get_neighbours(session=self.session)
        not_neighbours = Server.get_not_neighbours(session=self.session)

        changed_routes: t.Dict[Server, RouteContainer] = {}

        not_neighbours_anymore = []
        new_neighbours = []

        aws = []
        if check_current_neighbours:
            if neighbours:
                self.logger.debug(f"Checking current neighbours: " + ', '.join([str(s) for s in neighbours]))
                aws.append(_async_set_current_neighbours(neighbours, changed_routes))
            else:
                self.logger.debug(f"No neighbour to check")

        if discover_new_neighbours:
            if not_neighbours[:max_num_discovery]:
                rs = list(not_neighbours)
                random.shuffle(rs)
                target = rs[:max_num_discovery]
                target.sort(key=lambda s: s.name)
                self.logger.debug(
                    f"Checking new neighbours{f' (limited to {max_num_discovery})' if max_num_discovery else ''}: "
                    + ', '.join([str(s) for s in target]))
                aws.append(_async_discover_new_neighbours(target, changed_routes))
            else:
                self.logger.debug("No new neighbours to check")

        res = await asyncio.gather(*aws, return_exceptions=False)

        if check_current_neighbours and neighbours:
            not_neighbours_anymore = res.pop(0)
            if not_neighbours_anymore:
                self.logger.info(
                    f"Lost direct connection to the following nodes: " + ', '.join(
                        [str(s) for s in not_neighbours_anymore]))
        if discover_new_neighbours and not_neighbours[:max_num_discovery]:
            new_neighbours = res.pop(0)
            if new_neighbours:
                self.logger.info(f'New neighbours found: ' + ', '.join([str(s) for s in new_neighbours]))
            else:
                self.logger.debug("No new neighbours found")

        # remove routes whose proxy_server is a node that is not a neighbour
        query = self.session.query(Route).filter(
            Route.proxy_server_id.in_([s.id for s in list(set(not_neighbours).union(set(not_neighbours_anymore)))]))
        rc = RouteContainer(None, None, None)
        for route in query.all():
            route.set_route(rc)
            changed_routes[route.destination] = rc
        self.session.commit()

        # update neighbour lis

        neighbours = list(set(neighbours).union(set(new_neighbours)) - set(not_neighbours_anymore))

        if neighbours:
            self.logger.debug(f"Getting routing tables from {', '.join([str(s) for s in neighbours])}")
            responses = await asyncio.gather(
                *[ntwrk.async_get(server, 'api_1_0.routes', auth=get_root_auth()) for server in neighbours])

            cr = self._route_table_merge(dict(zip(neighbours, responses)))
            changed_routes.update(cr)

        return changed_routes

    def _update_route_table_from_data(self, new_routes: t.Dict, auth=None) -> t.Dict[Server, RouteContainer]:
        changed_routes = {}
        routes = new_routes.get('route_list', [])
        routes.sort(key=lambda x: x.get('cost') or MAX_COST, reverse=True)
        try:
            likely_proxy_server = self.session.query(Server).get(new_routes.get('server_id'))
            if not likely_proxy_server:
                self.logger.warning(f"Server id still '{new_routes.get('server_id')}' not created.")
                return changed_routes
            debug_new_routes = []
            for new_route in routes:
                target_server = self.session.query(Server).get(new_route.get('destination_id'))
                proxy_server = self.session.query(Server).get(new_route.get('proxy_server_id'))
                gate = self.session.query(Gate).get(new_route.get('gate_id'))
                dest_name = getattr(target_server, 'name', new_route.get('destination_id'))
                proxy_name = getattr(proxy_server, 'name', new_route.get('proxy_server_id'))
                gate_str = str(gate) if gate else new_route.get('gate_id')
                cost = new_route.get('cost')
                if gate_str and proxy_name:
                    gate_str = gate_str + '*' + proxy_name
                debug_new_routes.append(f'{dest_name} -> {gate_str or proxy_name} / {cost}')
                if target_server is None:
                    self.logger.warning(f"Destination server unknown {new_route.get('destination_id')}")
                    continue
                if target_server.id == self.server.id:
                    # check if server has detected me as a neighbour
                    if new_route.get('cost') == 0:
                        # check if I do not have it as a neighbour yet
                        if likely_proxy_server.route and likely_proxy_server.route.cost != 0:
                            # check if I have a gate to contact with it
                            route = check_gates(likely_proxy_server)
                            if isinstance(route, RouteContainer):
                                changed_routes[likely_proxy_server] = route
                                likely_proxy_server.set_route(route)
                else:
                    # server may be created without route (backward compatibility)
                    if target_server.route is None:
                        target_server.set_route(Route(destination=target_server))
                    # process routes whose proxy_server is not me
                    if self.server.id != new_route.get('proxy_server_id'):
                        # check If I reach destination
                        if target_server.route.cost is not None:
                            if new_route.get('cost') is None:
                                #  likely proxy does not reach but I reach it. It might be shutdown unexpectedly?
                                if target_server.route.cost == 0:
                                    # check if I still have it as a neighbour
                                    route = check_gates(target_server)
                                else:
                                    if target_server.route.proxy_server == likely_proxy_server:
                                        route = RouteContainer(None, None, None)
                                    else:
                                        # check if I still have access through my proxy
                                        cost, time = ntwrk.ping(target_server, retries=1, timeout=20,
                                                                session=self.session)
                                        if cost == target_server.route.cost:
                                            # still a valid route
                                            route = target_server.route
                                        elif cost is None:
                                            route = RouteContainer(None, None, None)
                                        else:
                                            route = RouteContainer(target_server.route.proxy_server, None, cost)

                                if isinstance(route, RouteContainer):
                                    # gate changed
                                    changed_routes[target_server] = route
                                    target_server.set_route(route)
                                elif route is None:
                                    # no route to host. I've lost contact too
                                    changed_routes[target_server] = RouteContainer(None, None, None)
                                    target_server.set_route(changed_routes[target_server])
                                else:
                                    # still a valid route. Send route to likely_proxy_server to tell it I have access
                                    resp = ntwrk.patch(likely_proxy_server, 'api_1_0.routes',
                                                           json=dict(server_id=str(self.server.id),
                                                                     route_list=[target_server.route.to_json()]),
                                                           auth=get_root_auth(), timeout=5)
                                    if not resp.ok:
                                        self.logger.info(f'Unable to send route to {likely_proxy_server}: {resp}')
                            elif target_server.route.proxy_server is not None and \
                                    target_server.route.proxy_server == likely_proxy_server:
                                # my proxy is telling me the route has changed
                                rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                                target_server.set_route(rc)
                                changed_routes.update({target_server: rc})
                            elif new_route.get('cost') + 1 < target_server.route.cost:
                                # if new route has less cost than actual route, take it as my new route
                                rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                                target_server.set_route(rc)
                                changed_routes.update({target_server: rc})
                        else:
                            # me does not reaches target_server
                            # if new route reaches the destination take it as a new one
                            if (new_route.get('cost') == 0 and new_route.get('gate_id') is not None) or (
                                    new_route.get('cost') is not None and proxy_server is not None):
                                rc = RouteContainer(likely_proxy_server, None, new_route.get('cost') + 1)
                            else:
                                # neither my route and the new route has access to the destination
                                rc = RouteContainer(None, None, None)
                            if target_server.route and (
                                    rc.proxy_server != target_server.route.proxy_server or
                                    rc.gate != target_server.route.gate or
                                    rc.cost != target_server.route.cost):
                                target_server.set_route(rc)
                                changed_routes.update({target_server: rc})
                    else:
                        # target_server reached through me as a proxy from likely_proxy
                        pass

            query = self.session.query(Route).filter(
                Route.proxy_server_id.in_([s.id for s, r in changed_routes.items() if r.cost is None]))
            rc = RouteContainer(None, None, None)
            for route in query.all():
                route.set_route(rc)
                changed_routes[route.destination] = rc

            self.logger.debug(
                f"New routes processed from {likely_proxy_server.name}: {json.dumps(debug_new_routes, indent=2)}")
            # if changed_routes:
            #     Parameter.set('routing_last_refresh', get_now())
        except errors.InvalidRoute as e:
            debug_new_routes = []
            routes.sort(key=lambda x: x.get('cost') or MAX_COST, reverse=True)
            for new_route in routes:
                target_server = self.session.query(Server).get(new_route.get('destination_id'))
                proxy_server = self.session.query(Server).get(new_route.get('proxy_server_id'))
                gate = self.session.query(Gate).get(new_route.get('gate_id'))
                dest_name = getattr(target_server, 'name', new_route.get('destination_id'))
                proxy_name = getattr(proxy_server, 'name', new_route.get('proxy_server_id'))
                gate_str = str(gate) if gate else new_route.get('gate_id')
                cost = new_route.get('cost')
                if gate_str and proxy_name:
                    gate_str = gate_str + '*' + proxy_name
                debug_new_routes.append(f'{dest_name} -> {gate_str or proxy_name} / {cost}')
            self.logger.exception(
                "Error setting routes from following data: " + json.dumps(debug_new_routes, indent=4))
        return changed_routes

    def _new_node_in_cluster(self, server_id, routes):
        changed_routes = {}
        server = self.session.query(Server).get(server_id)

        # server might be sent cluster in message but not created in database
        if server:
            new_route = check_gates(server, timeout=5, retries=3, delay=2)
            if new_route and isinstance(new_route, RouteContainer):
                self.logger.debug(f'cluster IN: New neighbour {server} found through {new_route.gate}')
                server.set_route(new_route)
                changed_routes.update({server: new_route})
            else:
                self.logger.debug(f"cluster IN: {server} is not a neighbour")
            changed_routes.update(
                self._update_route_table_from_data({'server_id': server_id, 'route_list': routes}))

        return changed_routes

    def _remove_node_from_cluster(self, server_id):
        changed_routes = {}
        server = self.session.query(Server).get(server_id)

        if server:
            server.set_route(RouteContainer(None, None, None))
            lost_routes = self.session.query(Route).filter_by(proxy_server_id=server.id).options(
                orm.lazyload(Route.destination), orm.lazyload(Route.gate), orm.lazyload(Route.proxy_server)).count()
            if lost_routes:
                changed_routes = self._loop.run_until_complete(
                    self._async_refresh_route_table(discover_new_neighbours=False,
                                                        check_current_neighbours=False))

            changed_routes.update({server: RouteContainer(None, None, None)})

        return changed_routes

    async def _send_routes(self, exclude=None):

        servers = Server.get_neighbours(session=self.session)
        msg, debug_msg = self._format_routes_message(self._changed_routes)

        c_exclude = []
        if self.logger.level <= logging.DEBUG:
            if exclude:
                if is_iterable_not_string(exclude):
                    c_exclude = [self.session.query(Server).get(e) if not isinstance(e, Server) else e for e in exclude]
                else:
                    c_exclude = [
                        self.session.query(Server).get(exclude) if not isinstance(exclude, Server) else exclude]
                log_msg = f" (Excluded nodes: {', '.join([getattr(e, 'name', e) for e in c_exclude])}):"
            else:
                log_msg = ''

            if servers:
                log_msg = f"Sending route information to the following nodes: {', '.join([s.name for s in servers])} " \
                          f"{log_msg}{json.dumps(debug_msg, indent=2)}"
            else:
                log_msg = f"No servers to send new routing information:{log_msg}{json.dumps(debug_msg, indent=2)}"
                if debug_msg:
                    log_msg += '\n' + json.dumps(debug_msg, indent=2)

            if debug_msg and (servers or exclude):
                self.logger.debug(log_msg)

        exclude_ids = list(set([s.id for s in servers]).union([getattr(e, 'id', e) for e in c_exclude]))

        auth = get_root_auth()
        aw = [ntwrk.async_patch(s, view_or_url='api_1_0.routes',
                                json={'server_id': self.server.id, 'route_list': msg,
                                      'exclude': exclude_ids},
                                auth=auth) for s in servers]

        rs = await asyncio.gather(*aw, return_exceptions=True)

        for r, s in zip(rs, servers):
            if isinstance(r, Exception):
                self.logger.warning(
                    f"Error while trying to send route data to node {s}: "
                    f"{format_exception(r)}")
            elif not r.ok:
                if r.exception:
                    self.logger.warning(
                        f"Error while trying to send route data to node {s}: "
                        f"{format_exception(r.exception)}")
                else:
                    self.logger.warning(f"Error while trying to send route data to node {s}: {r}")
        self._changed_routes.clear()

    def _format_routes_message(self, routes: t.Dict[t.Union[Id, Server], RouteContainer] = None) -> t.Tuple[
        t.List, t.List]:
        if routes:
            iterator = routes.items()
        else:
            iterator = [(r.destination, r) for r in self.session.query(Route).all()]
        msg = []
        debug_msg = []
        for d, r in iterator:
            msg.append(
                dict(destination_id=getattr(d, 'id', d),
                     proxy_server_id=str(getattr(r.proxy_server, 'id')) if getattr(r.proxy_server, 'id',
                                                                                   None) else None,
                     gate_id=str(getattr(r.gate, 'id')) if getattr(r.gate, 'id', None) else None,
                     cost=r.cost))

            dest_name = getattr(d, 'name', d)
            proxy_name = getattr(r.proxy_server, 'name') if getattr(r.proxy_server, 'name',
                                                                    None) else None
            gate_str = str(r.gate) if r.gate else None
            cost = r.cost
            if gate_str and proxy_name:
                gate_str = gate_str + '*' + proxy_name
            debug_msg.append(f'{dest_name} -> {gate_str or proxy_name} / {cost}')

        return msg, debug_msg
